from typing import Annotated, Generic, TypeVar

import numpy as np
from gpxpy.gpx import GPXTrackPoint
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PositiveInt,
    field_validator,
    model_validator,
)
from pydantic_numpy import np_array_pydantic_annotated_typing

from geo_track_analyzer.utils.internal import GPXTrackPointAfterValidator

T = TypeVar("T")


class Model(BaseModel):
    model_config = ConfigDict(extra="forbid")


class Position2D(Model):
    """Position in a 2D latitude / longitude space"""

    latitude: float
    """Latitude of the point"""

    longitude: float
    """Longitude of the point"""


class Position3D(Position2D):
    """Position in a 3D latitude / longitude / elevation space"""

    elevation: None | float
    """Elevation of the point"""


class ElevationMetrics(Model):
    """Collection of elevation related metrics"""

    uphill: float
    """Upill elevation in m"""

    downhill: float
    """Downhill elevation in m"""

    slopes: list[float]
    """Slopes between points in a uphill/downhill section"""


class SegmentOverviewMetric(Model, Generic[T]):
    """Collection max/average metric in a segement"""

    max: T
    """ Maximum value of the metric """

    avg: T
    """ Average value of the metric"""


class SegmentOverview(Model):
    """Collection of metrics for a Segment"""

    moving_time_seconds: float
    """Moving time of a segment"""

    total_time_seconds: float
    """Total time of the segment"""

    moving_distance: float
    """Moving distance (point-to-point distance with velocity below a threshold) in m"""

    total_distance: float
    """Total distance of the segment in m"""

    velocity: SegmentOverviewMetric[float] | None
    """Velocity in the segment in m/s (only considering velocities below the XX percentile). Saved as [`SegmentOverviewMetric`][geo_track_analyzer.model.SegmentOverviewMetric]"""  # noqa: E501

    max_elevation: None | float
    """Maximum elevation in the segment in m"""

    min_elevation: None | float
    """ Minimum elevation in the segment in m"""

    uphill_elevation: None | float
    """Elevation traveled uphill in m"""

    downhill_elevation: None | float
    """Elevation traveled downhill in m"""

    heartrate: SegmentOverviewMetric[int] | None = Field(default=None)
    """ Heartrate metrics in bpm. Saved as [`SegmentOverviewMetric`][geo_track_analyzer.model.SegmentOverviewMetric] """  # noqa: E501

    power: SegmentOverviewMetric[int] | None = Field(default=None)
    """ Power metrics in watts. Saved as [`SegmentOverviewMetric`][geo_track_analyzer.model.SegmentOverviewMetric] """  # noqa: E501

    cadence: SegmentOverviewMetric[int] | None = Field(default=None)
    """ Cadence metrics in rpm. Saved as [`SegmentOverviewMetric`][geo_track_analyzer.model.SegmentOverviewMetric] """  # noqa: E501

    # Attributes that will be calculated from primary attributes
    moving_distance_km: float = Field(default=-1)
    """moving_distance converted the km"""

    total_distance_km: float = Field(default=-1)
    """total_distance converted the km"""

    velocity_kmh: SegmentOverviewMetric[None | float] | None = Field(default=None)
    """ Velocity in the segment in km/h """

    @model_validator(mode="after")  # type: ignore
    def set_km_attr(self) -> "SegmentOverview":
        self.moving_distance_km = self.moving_distance / 1000
        self.total_distance_km = self.total_distance / 1000
        self.velocity_kmh = (
            SegmentOverviewMetric(
                max=3.6 * self.velocity.max,
                avg=3.6 * self.velocity.avg,
            )
            if self.velocity is not None
            else None
        )
        return self


class SegmentOverlap(Model):
    """Represent the overlap between two segments"""

    overlap: float
    """Overlap ratio of the segments"""

    inverse: bool
    """Match direction of the segment relative to the base"""

    plate: np_array_pydantic_annotated_typing(data_type=np.float32, dimensions=2)  # type: ignore
    """2D representation of the segment overlap"""

    start_point: Annotated[GPXTrackPoint, GPXTrackPointAfterValidator]
    """First point matching the base segment """

    start_idx: int
    """Index of the first point in match segment"""

    end_point: Annotated[GPXTrackPoint, GPXTrackPointAfterValidator]
    """Last point matching the base segment """

    end_idx: int
    """Index of the last point in match segment"""

    def __repr__(self) -> str:
        ret_str = f"Overlap {self.overlap * 100:.2f}%, Inverse: {self.inverse},"
        ret_str += f" Plate: {self.plate.shape}, Points: "
        point_strs = []
        for point, idx in zip(
            [self.start_point, self.end_point], [self.start_idx, self.end_idx]
        ):
            point_strs.append(f"({point.latitude},{point.longitude}) at id {idx}")
        ret_str += " to ".join(point_strs)

        return ret_str


class PointDistance(Model):
    """Represents a distance calculati  on to a point in a GPX track."""

    point: Annotated[GPXTrackPoint, GPXTrackPointAfterValidator]
    """The nearest point on the track."""

    distance: float
    """The distance to the nearest point."""

    point_idx_abs: int
    """The absolute index of the nearest point in the track."""

    segment_idx: int
    """The index of the segment containing the nearest point."""

    segment_point_idx: int
    """The index of the nearest point within the containing segment."""

    def __repr__(self) -> str:
        return (
            f"PointDistance(point=({self.point.latitude},{self.point.longitude})"
            f"distance={self.distance}, "
            f"point_idx_abs={self.point_idx_abs}, "
            f"segment_idx={self.segment_idx}, "
            f"segment_point_idx={self.segment_point_idx})"
        )


class ZoneInterval(Model):
    """Represents a zone interval. Start or end can be None implying and interval that
    include all value smaller or larger than the provided value, respectively

    :raises ValueError: If start and end are None
    """

    start: None | PositiveInt
    """Lower bound of the interval"""

    end: None | PositiveInt
    """Upper bound of the interval"""

    name: None | str = None
    """Optional name of the interval"""

    color: None | str = None
    """Optional color of the interval"""

    # TODO: Add color validation
    @model_validator(mode="after")  # type: ignore
    def check_zone_is_valid(self) -> "ZoneInterval":
        if self.start is None and self.end is None:
            raise ValueError("start and end can not both be None")
        return self


class Zones(Model):
    """Represents a collection of zones. The list of intervals is required to contain
    open ZoneIntervals in the fist and last position. Furthermore the intervals can not
    have gaps.

    :raises ValueError: If not at least two intervals are provided in the object
    :raises ValueError: If some intervals have a name attached but not all (or none)
    :raises ValueError: If some intervals have a color attached but not all (or none)
    :raises ValueError: If first interval has a start value not equal to None
    :raises ValueError: If last interval has a end value not equal to None
    :raises ValueError: If consecutive intervals to not end/start with the same value
    """

    intervals: list[ZoneInterval]
    """A list of intervals comprising the collection of zones"""

    meta_info: None | str = None
    """Optional meta information about the zones"""

    @field_validator("intervals")
    @classmethod
    def at_least_two_intervals(cls, v: list[ZoneInterval]) -> list:
        if len(v) < 2:
            raise ValueError("At least two intervals are required")

        n_none_names = 0
        n_none_colors = 0
        for interval in v:
            if interval.name is None:
                n_none_names += 1
            if interval.color is None:
                n_none_colors += 1

        if n_none_names > 0 and n_none_names != len(v):
            raise ValueError("Set either no names of intervals or all names")

        if n_none_colors > 0 and n_none_colors != len(v):
            raise ValueError("Set either no color of intervals or all colors")

        return v

    @model_validator(mode="after")  # type: ignore
    def check_intervals(self) -> "Zones":
        if self.intervals[0].start is not None:
            raise ValueError("First interval must start with None")
        if self.intervals[-1].end is not None:
            raise ValueError("Last interval must end with None")
        prev_interval = self.intervals[0]
        for interval in self.intervals[1:]:
            if interval.start != prev_interval.end:
                raise ValueError(
                    "Consecutive intervals mit start/end with the same value"
                )
            prev_interval = interval
        return self



import click

from geo_track_analyzer.track import FITTrack


@click.command()
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "--name", help="Optinally set the Name property in the output file", default=None
)
@click.option(
    "--email", help="Optinally set the eMail property in the output file", default=None
)
def main(filename: str, name: None | str, email: None | str) -> None:
    """
    Extract the track information for FILENAME file in fit format and
    save to regular gpx file
    """
    click.echo("Extracting track from %s" % filename)
    track = FITTrack(filename)

    out_file_name = filename.replace(".fit", "") + ".gpx"
    click.echo("Writing file %s" % out_file_name)
    with open(out_file_name, "w") as f:
        f.write(track.get_xml(name=name, email=email))


if __name__ == "__main__":
    main()



from typing import Any, Literal

import pandas as pd
import plotly.graph_objects as go
from plotly.colors import make_colorscale, sample_colorscale
from plotly.graph_objs import Figure

from geo_track_analyzer.exceptions import VisualizationSetupError
from geo_track_analyzer.visualize.constants import (
    COLOR_GRADIENTS,
    DEFAULT_BAR_COLORS,
    DEFAULT_COLOR_GRADIENT,
    ENRICH_UNITS,
)
from geo_track_analyzer.visualize.utils import get_color_gradient


def _preprocess_data(
    data: pd.DataFrame,
    metric: Literal["heartrate", "power", "cadence"],
    strict_data_selection: bool,
) -> pd.DataFrame:
    if metric not in data.columns:
        raise VisualizationSetupError("Metric %s not part of the passed data" % metric)

    mask = data.moving
    if strict_data_selection:
        mask = mask & data.in_speed_percentile

    data_for_plot = data[mask]

    if f"{metric}_zones" not in data_for_plot.columns:
        raise VisualizationSetupError("Zone data is not provided in passed dataframe")

    if pd.isna(data_for_plot[metric]).all():
        raise VisualizationSetupError(
            "Requested to plot %s but information is not available in data" % metric
        )

    return data_for_plot


def _aggregate_zone_data(
    data: pd.DataFrame,
    metric: Literal["heartrate", "power", "cadence"],
    aggregate: Literal["time", "distance", "speed"],
    aggregation_method: Literal["sum", "mean"],
    time_as_timedelta: bool = False,
) -> tuple[pd.DataFrame, str, str]:
    group = data.groupby(f"{metric}_zones")
    # Make sure that the groups are ordered by metric value
    bin_data = (
        pd.concat(
            [group[aggregate].agg(aggregation_method), group[metric].min()],  # type: ignore
            axis=1,
        )
        .sort_values(metric)[aggregate]
        .reset_index()
    )
    if aggregate == "time":
        bin_data["time"] = (
            pd.to_timedelta(bin_data["time"].astype(int), unit="s")
            if time_as_timedelta
            else pd.to_datetime(bin_data["time"].astype(int), unit="s")
        )
        y_title = "duration"
        tickformat = "%H:%M:%S"
    elif aggregate == "distance":
        bin_data["distance"] = bin_data["distance"] / 1000
        y_title = "distance [km]"
        tickformat = ""
    elif aggregate == "speed":
        bin_data["speed"] = bin_data["speed"] * 3.6
        y_title = "velocity [km/h]"
        tickformat = ""
    else:
        raise NotImplementedError(f"Aggregation {aggregate} not supported")

    if aggregation_method == "sum":
        y_title = f"Total {y_title}"
    elif aggregation_method == "mean":
        y_title = f"Average {y_title}"

    bin_data["colors"] = group[f"{metric}_zone_colors"].first().to_numpy()

    return bin_data, y_title, tickformat


def plot_track_zones(
    data: pd.DataFrame,
    metric: Literal["heartrate", "power", "cadence"],
    aggregate: Literal["time", "distance", "speed"],
    *,
    use_zone_colors: bool = False,
    height: None | int = 600,
    width: None | int = 1200,
    strict_data_selection: bool = False,
    as_pie_chart: bool = False,
) -> Figure:
    """Aggregate a value per zone defined for heartrate, power, or cadence.

    :param data: DataFrame containing track and zone data
    :param metric: One of "heartrate", "cadence", or "power"
    :param aggregate: Value to aggregate. Supported values are (total) "time",
        "distance",  and (average) speed in a certain zone
    :param use_zone_colors: If True, use distinct colors per zone (either set by the
        zone object or a default defined by the package). Otherwise alternating colors
        will be used, defaults to False.
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1200
    :param strict_data_selection: If True only included that passing the minimum speed
        requirements of the Track, defaults to False
    :raises VisualizationSetupError: Is raised if metric is not avaialable in the data

    :return: Plotly Figure object
    """
    data_for_plot = _preprocess_data(data, metric, strict_data_selection)

    bin_data, y_title, tickformat = _aggregate_zone_data(
        data_for_plot,
        metric,
        aggregate,
        aggregation_method="mean" if aggregate == "speed" else "sum",
        time_as_timedelta=as_pie_chart,
    )

    if use_zone_colors:
        colors = bin_data.colors
    else:
        if as_pie_chart:
            col_a, col_b = COLOR_GRADIENTS.get(metric, DEFAULT_COLOR_GRADIENT)
            colors = get_color_gradient(col_a, col_b, len(bin_data))
        else:
            col_a, col_b = DEFAULT_BAR_COLORS
            colors = []
            for i in range(len(bin_data)):
                colors.append(col_a if i % 2 == 0 else col_b)

    if as_pie_chart:
        unit = ENRICH_UNITS.get(aggregate, "")
        hover_template = "{value:.2f} {unit}"
        if aggregate == "time":
            hover_template = "{value} {unit} "
        fig = go.Figure(
            go.Pie(
                labels=bin_data[f"{metric}_zones"],
                values=bin_data[aggregate],
                marker=dict(colors=colors, line=dict(color="#000000", width=1)),
                hovertemplate="%{hovertext}<extra></extra>",
                hovertext=[
                    hover_template.format(value=v, unit=unit)
                    for v in bin_data[aggregate]
                ],
                sort=False,
                direction="clockwise",
                hole=0.3,
                textinfo="label+percent",
                textposition="outside",
            )
        )
        fig.update_layout(showlegend=False)
    else:
        fig = go.Figure(
            go.Bar(
                x=bin_data[f"{metric}_zones"],
                y=bin_data[aggregate],
                marker_color=colors,
                hoverinfo="skip",
            ),
        )

    for i, rcrd in enumerate(bin_data.to_dict("records")):
        if as_pie_chart:
            continue
        kwargs: dict[str, Any] = dict(
            x=i,
            showarrow=False,
            yshift=10,
        )
        if aggregate == "time":
            kwargs.update(
                dict(
                    y=rcrd["time"],
                    text=rcrd["time"].time().isoformat(),
                )
            )
        elif aggregate == "distance":
            kwargs.update(
                dict(
                    y=rcrd["distance"],
                    text=f"{rcrd['distance']:.2f} km",
                )
            )
        elif aggregate == "speed":
            kwargs.update(
                dict(
                    y=rcrd["speed"],
                    text=f"{rcrd['speed']:.2f} km/h",
                )
            )
        else:
            raise NotImplementedError(f"Aggregate {aggregate} is not implemented")

        fig.add_annotation(**kwargs)

    fig.update_layout(
        title=f"{aggregate.capitalize()} in {metric.capitalize()} zones",
        yaxis=dict(tickformat=tickformat, title=y_title),
        bargap=0.0,
    )

    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    return fig


def plot_segment_zones(
    data: pd.DataFrame,
    metric: Literal["heartrate", "power", "cadence"],
    aggregate: Literal["time", "distance", "speed"],
    *,
    bar_colors: None | tuple[str, str] | list[str] = None,
    height: None | int = 600,
    width: None | int = 1200,
    strict_data_selection: bool = False,
) -> Figure:
    """Aggregate a value per zone defined for heartrate, power, or cadence, split into
    segments available in data.

    :param data: DataFrame containing track and zone data
    :param metric: One of heartrate, cadence, or power
    :param aggregate: Value to aggregate. Supported values are (total) "time",
        "distance",  and (average) speed in a certain zone
    :param bar_colors: Overwrite the default colors for the bar. If a tuple of two
        colors is passed, a colorscale will be generated based on these values and
        colors for segments will be picked from this scale. Furthermore, a list of
        colors can be passed that must at least be as long as the number of segments in
        the data
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1200
    :param strict_data_selection: If True only included that passing the minimum speed
        requirements of the Track, defaults to False
    :raises VisualizationSetupError: Is raised if metric is not avaialable in the data
    :raises VisualizationSetupError: Is raised if no segment information is available in
        the data

    :return: Plotly Figure object
    """
    if "segment" not in data.columns:
        raise VisualizationSetupError(
            "Data has no **segment** in columns. Required for plot"
        )
    data_for_plot = _preprocess_data(data, metric, strict_data_selection)

    fig = go.Figure()

    plot_segments = data_for_plot.segment.unique()

    if isinstance(bar_colors, list):
        if len(plot_segments) > len(bar_colors):
            raise VisualizationSetupError(
                "If a list of colors is passed, it must be at least same lenght as the "
                "segments in the data"
            )
        colors = bar_colors[0 : len(plot_segments)]
    else:
        colors = sample_colorscale(
            make_colorscale(DEFAULT_BAR_COLORS if bar_colors is None else bar_colors),
            len(plot_segments),
        )

    for color, segment in zip(colors, plot_segments):
        _data_for_plot = data_for_plot[data_for_plot.segment == segment]
        bin_data, y_title, tickformat = _aggregate_zone_data(
            _data_for_plot,
            metric,
            aggregate,
            aggregation_method="mean" if aggregate == "speed" else "sum",
        )

        hovertext = []
        for rcrd in bin_data.to_dict("records"):
            if aggregate == "time":
                hovertext.append(rcrd["time"].time().isoformat())

            elif aggregate == "distance":
                hovertext.append(f"{rcrd['distance']:.2f} km")

            elif aggregate == "speed":
                hovertext.append(f"{rcrd['speed']:.2f} km/h")

        fig.add_trace(
            go.Bar(
                x=bin_data[f"{metric}_zones"],
                y=bin_data[aggregate],
                name=f"Segment {segment}",
                marker_color=color,
                hovertext=hovertext,
                hovertemplate="%{hovertext}<extra></extra>",
            ),
        )

    fig.update_layout(
        title=f"{aggregate.capitalize()} in {metric.capitalize()} zones",
        yaxis=dict(tickformat=tickformat, title=y_title),
    )

    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    return fig


def plot_segment_summary(
    data: pd.DataFrame,
    aggregate: Literal["total_time", "total_distance", "avg_speed", "max_speed"],
    *,
    colors: None | tuple[str, str] = None,
    height: None | int = 600,
    width: None | int = 1200,
    strict_data_selection: bool = False,
) -> Figure:
    """_summary_

    :param data: DataFrame containing track and zone data
    :param aggregate: Value to aggregate. Supported values are "total_time",
        "total_distance", "avg_speed", and "max_speed"
    :param colors: Overwrite the default alternating colors, defaults to None
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1200
    :param strict_data_selection: If True only included that passing the minimum speed
        requirements of the Track, defaults to False
    :raises VisualizationSetupError: Is raised if no segment information is available in
        the data

    :return: Plotly Figure object
    """
    if "segment" not in data.columns:
        raise VisualizationSetupError(
            "Data has no **segment** in columns. Required for plot"
        )

    if colors is None:
        colors = DEFAULT_BAR_COLORS
    col_a, col_b = colors

    mask = data.moving
    if strict_data_selection:
        mask = mask & data.in_speed_percentile

    _data_for_plot = data[mask]

    fig = go.Figure()

    if aggregate == "avg_speed":
        bin_data = _data_for_plot.groupby("segment").speed.agg("mean") * 3.6
        y_title = "Average velocity [km/h]"
        tickformat = ""
        hover_map_func = lambda v: str(f"{v:.2f} km/h")
    elif aggregate == "max_speed":
        bin_data = _data_for_plot.groupby("segment").speed.agg("max") * 3.6
        y_title = "Maximum velocity [km/h]"
        tickformat = ""
        hover_map_func = lambda v: str(f"{v:.2f} km/h")
    elif aggregate == "total_distance":
        bin_data = _data_for_plot.groupby("segment").distance.agg("sum") / 1000
        y_title = "Distance [km]"
        tickformat = ""
        hover_map_func = lambda v: str(f"{v:.2f} km")
    elif aggregate == "total_time":
        bin_data = pd.to_datetime(
            _data_for_plot.groupby("segment").time.agg("sum"), unit="s"
        )
        y_title = "Duration"
        tickformat = "%H:%M:%S"
        hover_map_func = lambda dt: str(dt.time())
    else:
        raise NotImplementedError(f"Aggregate {aggregate} is not implemented")

    fig.add_trace(
        go.Bar(
            x=[f"Segment {idx}" for idx in bin_data.index.to_list()],
            y=bin_data.to_list(),
            marker_color=[col_a if i % 2 == 0 else col_b for i in range(len(bin_data))],
            hovertext=list(map(hover_map_func, bin_data.to_list())),
            hovertemplate="%{hovertext}<extra></extra>",
        ),
    )

    fig.update_layout(
        yaxis=dict(tickformat=tickformat, title=y_title),
        bargap=0.0,
    )

    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    return fig


def plot_segment_box_summary(
    data: pd.DataFrame,
    metric: Literal["heartrate", "power", "cadence", "speed", "elevation"],
    *,
    colors: None | tuple[str, str] = None,
    height: None | int = 600,
    width: None | int = 1200,
    strict_data_selection: bool = False,
) -> Figure:
    """Show the metric as boxplot for each segment in the data.

    :param data: DataFrame containing track and zone data
    :param metric: One of "heartrate", "cadence", "power", or "speed"
    :param colors: Overwrite the default alternating colors, defaults to None
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1200
    :param strict_data_selection: If True only included that passing the minimum speed
        requirements of the Track, defaults to False
    :raises VisualizationSetupError: Is raised if metric is not avaialable in the data
    :raises VisualizationSetupError: Is raised if no segment information is available in
        the data

    :return: Plotly Figure object
    """
    if "segment" not in data.columns:
        raise VisualizationSetupError(
            "Data has no **segment** in columns. Required for plot"
        )

    if metric not in data.columns:
        raise VisualizationSetupError("Metric %s not part of the passed data" % metric)

    if colors is None:
        colors = DEFAULT_BAR_COLORS
    col_a, col_b = colors

    mask = data.moving
    if strict_data_selection:
        mask = mask & data.in_speed_percentile

    data_for_plot = data[mask]

    fig = go.Figure()

    for i, segment in enumerate(data_for_plot.segment.unique()):
        _data_for_plot = data_for_plot[data_for_plot.segment == segment]

        if metric == "speed":
            box_data = _data_for_plot["speed"] * 3.6
        else:
            box_data = _data_for_plot[metric]

        fig.add_trace(
            go.Box(
                y=box_data,
                name=f"Segment {segment}",
                boxpoints=False,
                line_color=col_a if i % 2 == 0 else col_b,
                marker_color=col_a if i % 2 == 0 else col_b,
            )
        )
    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    fig.update_layout(
        yaxis=dict(title=f"{metric.capitalize()} {ENRICH_UNITS[metric]}"),
        showlegend=False,
    )

    return fig



def generate_error_text(entry_point: str) -> str:
    return (
        f"The {entry_point} command line interface can not be run because of missing "
        "dependencies. Make sure that the package was installed with the **cli** "
        "extra. Use: pip install geo-track-analyzer[cli]"
    )


try:
    from ._update_elevation import main as update_elevation
except ImportError:

    def update_elevation() -> None:
        import sys

        print(generate_error_text("enhance-elevation"))
        sys.exit(1)


try:
    from ._extract_track import main as extract_track
except ImportError:

    def extract_track() -> None:
        import sys

        print(generate_error_text("extract-fit-track"))
        sys.exit(1)


__all__ = ["update_elevation", "extract_track"]



import logging
from datetime import timedelta
from math import acos, asin, atan2, cos, degrees, pi, sin, sqrt
from typing import Callable, Literal, Type, TypeVar, Union

import numpy as np
import numpy.typing as npt
from gpxpy.gpx import GPXBounds, GPXTrack, GPXTrackPoint, GPXTrackSegment

from geo_track_analyzer.exceptions import (
    GPXPointExtensionError,
    InvalidBoundsError,
    TrackAnalysisError,
)
from geo_track_analyzer.model import (
    ElevationMetrics,
    PointDistance,
    Position2D,
    Position3D,
)
from geo_track_analyzer.utils.internal import (
    get_extended_track_point,
    get_extension_value,
)

try:
    import coloredlogs  # type: ignore
except ModuleNotFoundError:
    coloredlogs = None


logger = logging.getLogger(__name__)

T = TypeVar("T", float, int)


def distance(pos1: Position2D, pos2: Position2D) -> float:
    """
    Calculate the distance between to long/lat points using the Haversine formula.
    Following: https://stackoverflow.com/a/21623206

    :param pos1: Lat/long position 1
    :param pos2: Lat/long position 2

    :returns: Distance in m
    """
    p = pi / 180
    a = (
        0.5
        - cos((pos2.latitude - pos1.latitude) * p) / 2
        + cos(pos1.latitude * p)
        * cos(pos2.latitude * p)
        * (1 - cos((pos2.longitude - pos1.longitude) * p))
        / 2
    )
    distance_km = 12742 * asin(sqrt(a))
    return distance_km * 1000


def get_latitude_at_distance(
    position: Position2D, distance: float, to_east: bool
) -> float:
    a = pow(sin(distance / 12742000), 2)
    b = acos(1 - 2 * a) / (pi / 180)
    if to_east:
        return b + position.latitude
    return position.latitude - b


def get_longitude_at_distance(
    position: Position2D, distance: float, to_north: bool
) -> float:
    p = pi / 180
    a = pow(sin(distance / 12742000), 2)
    b = pow(cos(position.latitude * p), 2) / 2
    c = acos(1 - (a / b)) / p

    if to_north:
        return c + position.longitude
    return position.longitude - c


def calc_elevation_metrics(
    positions: list[Position3D],
) -> ElevationMetrics:
    """
    Calculate elevation related metrics for the passed list of Position3D objects

    :param positions: Position3D object containing latitude, longitude and elevation

    :returns: A ElevationMetrics object containing uphill and downhill distances and the
        point-to-point slopes.
    """
    uphill = 0.0
    downhill = 0.0
    slopes = [0.0]  # Pad with slope 0 so len(slopes) == len(positions)
    for prev_pos, curr_pos in zip(positions, positions[1::]):  # noqa: RUF007
        if curr_pos == prev_pos:
            continue
        if curr_pos.elevation is None or prev_pos.elevation is None:
            continue

        pp_elevation = curr_pos.elevation - prev_pos.elevation
        pp_distance = distance(prev_pos, curr_pos)

        if pp_elevation > 0:
            uphill += pp_elevation
        else:
            downhill += pp_elevation
        try:
            o_by_h = pp_elevation / pp_distance
        except ZeroDivisionError:
            o_by_h = 0
        # Addressing **ValueError: math domain error**
        if o_by_h > 1 or o_by_h < -1:
            slopes.append(np.nan)
        else:
            slopes.append(degrees(asin(o_by_h)))

    return ElevationMetrics(uphill=uphill, downhill=abs(downhill), slopes=slopes)


def parse_level(this_level: Union[int, str]) -> tuple[int, Callable]:
    if this_level == 20 or this_level == "INFO":
        return logging.INFO, logging.info
    elif this_level == 10 or this_level == "DEBUG":
        return logging.DEBUG, logging.debug
    elif this_level == 30 or this_level == "WARNING":
        return logging.WARNING, logging.warning
    elif this_level == 40 or this_level == "ERROR":
        return logging.ERROR, logging.error
    elif this_level == 50 or this_level == "CRITICAL":
        return logging.CRITICAL, logging.critical
    else:
        raise RuntimeError("%s is not supported" % this_level)


def init_logging(this_level: Union[int, str]) -> bool:
    """Helper function for setting up python logging"""
    log_format = "[%(asctime)s] %(name)-30s %(levelname)-8s %(message)s"
    level, _ = parse_level(this_level)
    if coloredlogs is None:
        logging.basicConfig(level=level, format=log_format)
    else:
        coloredlogs.install(level=level, fmt=log_format)
    return True


def center_geolocation(geolocations: list[tuple[float, float]]) -> tuple[float, float]:
    """
    Calculate an estimated (based on the assumption the earth is a perfect sphere) given
    a list of latitude, longitude pairs in degree.

    Based on: https://gist.github.com/amites/3718961

    :param geolocations: list of latitude, longitude pairs in degree

    :returns: Estimate center latitude, longitude pair in degree
    """
    x, y, z = 0.0, 0.0, 0.0

    for lat, lon in geolocations:
        lat = float(lat) * pi / 180
        lon = float(lon) * pi / 180
        x += cos(lat) * cos(lon)
        y += cos(lat) * sin(lon)
        z += sin(lat)

    x = float(x / len(geolocations))
    y = float(y / len(geolocations))
    z = float(z / len(geolocations))

    lat_c, lon_c = atan2(z, sqrt(x * x + y * y)), atan2(y, x)

    return lat_c * 180 / pi, lon_c * 180 / pi


def interpolate_linear(start_value: T, end_val: T, points: int) -> list[float]:
    fracs = np.arange(0, points + 1)
    return np.interp(fracs, [0, points], [start_value, end_val]).tolist()


def interpolate_extension(
    start: GPXTrackPoint,
    end: GPXTrackPoint,
    extension: str,
    n_points: int,
    interpolation_type: Literal["copy-forward", "meet-center", "linear"],
    convert_type: Type[T],
) -> list[None] | list[T]:
    try:
        value_start = convert_type(get_extension_value(start, extension))
    except GPXPointExtensionError:
        value_start = None
    try:
        value_end = convert_type(get_extension_value(end, extension))
    except GPXPointExtensionError:
        value_end = None

    if value_start is None or value_end is None:
        return [None for _ in range(n_points)]

    if interpolation_type == "linear":
        return [
            convert_type(v)
            for v in interpolate_linear(value_start, value_end, n_points - 1)
        ]
    elif interpolation_type == "copy-forward":
        return [value_start for _ in range(n_points - 1)] + [value_end]
    else:
        n_second_half = n_points // 2
        n_first_half = n_points - n_second_half
        return [value_start for _ in range(n_first_half)] + [
            value_end for _ in range(n_second_half)
        ]


def interpolate_points(
    start: GPXTrackPoint,
    end: GPXTrackPoint,
    spacing: float,
    copy_extensions: Literal["copy-forward", "meet-center", "linear"] = "copy-forward",
) -> None | list[GPXTrackPoint]:
    """
    Simple linear interpolation between GPXTrackPoint. Supports latitude, longitude
    (required), elevation (optional), and time (optional)
    """

    pp_distance = distance(
        Position2D(latitude=start.latitude, longitude=start.longitude),
        Position2D(latitude=end.latitude, longitude=end.longitude),
    )
    if pp_distance < 2 * spacing:
        return None

    logger.debug(
        "pp-distance %s | n_points interpol %s ", pp_distance, pp_distance // spacing
    )

    n_points: int = round(pp_distance // spacing)

    lat_int = interpolate_linear(start.latitude, end.latitude, n_points)
    lng_int = interpolate_linear(start.longitude, end.longitude, n_points)

    if start.elevation is None or end.elevation is None:
        elevation_int = len(lng_int) * [None]
    else:
        elevation_int = interpolate_linear(start.elevation, end.elevation, n_points)

    if start.time is None or end.time is None:
        time_int = len(lng_int) * [None]
    else:
        time_int = interpolate_linear(
            0, (end.time - start.time).total_seconds(), n_points
        )

    hr_int = interpolate_extension(
        start, end, "heartrate", n_points + 1, copy_extensions, int
    )
    cd_int = interpolate_extension(
        start, end, "cadence", n_points + 1, copy_extensions, int
    )
    pw_int = interpolate_extension(
        start, end, "power", n_points + 1, copy_extensions, int
    )

    ret_points = []

    for i in range(len(lat_int)):
        if time_int[i] is not None:
            time = start.time + timedelta(seconds=time_int[i])
        else:
            time = None

        this_extensions = {}
        if hr_int[i] is not None:
            this_extensions["heartrate"] = hr_int[i]
        if cd_int[i] is not None:
            this_extensions["cadence"] = cd_int[i]
        if pw_int[i] is not None:
            this_extensions["power"] = pw_int[i]

        ret_points.append(
            get_extended_track_point(
                lat=lat_int[i],
                lng=lng_int[i],
                ele=elevation_int[i],
                timestamp=time,
                extensions=this_extensions,
            )
        )
        logger.debug(
            "New point %s / %s / %s / %s -> distance to origin %s",
            lat_int[i],
            lng_int[i],
            elevation_int[i],
            time,
            distance(
                Position2D(latitude=start.latitude, longitude=start.longitude),
                Position2D(latitude=lat_int[i], longitude=lng_int[i]),
            ),
        )

    return ret_points


def interpolate_segment(
    segment: GPXTrackSegment,
    spacing: float,
    copy_extensions: Literal["copy-forward", "meet-center", "linear"] = "copy-forward",
) -> GPXTrackSegment:
    """
    Interpolate points in a GPXTrackSegment to achieve a specified spacing.

    :param segment: GPXTrackSegment to interpolate.
    :param spacing: Desired spacing between interpolated points.
    :param copy_extension: How should the extenstion (if present) be defined in the
        interpolated points.
    :return: Interpolated GPXTrackSegment with points spaced according to the specified
        spacing.
    """
    init_points = segment.points

    new_segment_points = []
    for i, (start, end) in enumerate(
        zip(init_points[:-1], init_points[1:])  # noqa: RUF007
    ):
        new_points = interpolate_points(
            start=start,
            end=end,
            spacing=spacing,
            copy_extensions=copy_extensions,
        )

        if new_points is None:
            if i == 0:
                new_segment_points.extend([start, end])
            else:
                new_segment_points.extend([end])
            continue

        if i == 0:
            new_segment_points.extend(new_points)
        else:
            new_segment_points.extend(new_points[1:])

    interpolated_segment = GPXTrackSegment()
    interpolated_segment.points = new_segment_points
    return interpolated_segment


def get_segment_base_area(segment: GPXTrackSegment) -> float:
    """Caculate the area enclodes by the bounds in m^2"""
    bounds = segment.get_bounds()

    try:
        check_bounds(bounds)
    except InvalidBoundsError:
        return 0

    # After check_bounds this always works
    latitude_distance = distance(
        Position2D(latitude=bounds.max_latitude, longitude=bounds.min_longitude),  # type: ignore
        Position2D(latitude=bounds.min_latitude, longitude=bounds.min_longitude),  # type: ignore
    )

    longitude_distance = distance(
        Position2D(latitude=bounds.min_latitude, longitude=bounds.max_longitude),  # type: ignore
        Position2D(latitude=bounds.min_latitude, longitude=bounds.min_longitude),  # type: ignore
    )

    return latitude_distance * longitude_distance


def crop_segment_to_bounds(
    segment: GPXTrackSegment,
    bounds_min_latitude: float,
    bounds_min_longitude: float,
    bounds_max_latitude: float,
    bounds_max_longitude: float,
) -> GPXTrackSegment:
    """
    Crop a GPXTrackSegment to include only points within specified geographical bounds.

    :param segment: GPXTrackSegment to be cropped.
    :param bounds_min_latitude: Minimum latitude of the geographical bounds.
    :param bounds_min_longitude: Minimum longitude of the geographical bounds.
    :param bounds_max_latitude: Maximum latitude of the geographical bounds.
    :param bounds_max_longitude: Maximum longitude of the geographical bounds.

    :return: Cropped GPXTrackSegment containing only points within the specified bounds.
    """
    cropped_segment = GPXTrackSegment()
    for point in segment.points:
        if (bounds_min_latitude <= point.latitude <= bounds_max_latitude) and (
            bounds_min_longitude <= point.longitude <= bounds_max_longitude
        ):
            cropped_segment.points.append(point)

    return cropped_segment


def get_distances(v1: npt.NDArray, v2: npt.NDArray) -> npt.NDArray:
    """
    Calculates the distances between two sets of latitude/longitude pairs.

    :param v1: A NumPy array of shape (N, 2) containing latitude/longitude pairs.
    :param v2: A NumPy array of shape (N, 2) containing latitude/longitude pairs.

    :return: A NumPy array of shape (N, M) containing the distances between the
        corresponding pairs in v1 and v2.
    """
    v1_lats, v1_longs = v1[:, 0], v1[:, 1]
    v2_lats, v2_longs = v2[:, 0], v2[:, 1]

    v1_lats = np.reshape(v1_lats, (v1_lats.shape[0], 1))
    v2_lats = np.reshape(v2_lats, (1, v2_lats.shape[0]))

    v1_longs = np.reshape(v1_longs, (v1_longs.shape[0], 1))
    v2_longs = np.reshape(v2_longs, (1, v2_longs.shape[0]))

    # pi vec
    v_pi = np.reshape(np.ones(v1_lats.shape[0]) * (pi / 180), (v1_lats.shape[0], 1))

    dp = (
        0.5
        - np.cos((v2_lats - v1_lats) * v_pi) / 2  # type: ignore
        + np.cos(v1_lats * v_pi)  # type: ignore
        * np.cos(v2_lats * v_pi)  # type: ignore
        * (1 - np.cos((v2_longs - v1_longs) * v_pi))  # type: ignore
        / 2
    )

    dp_km = 12742 * np.arcsin(np.sqrt(dp))

    return dp_km * 1000


def distance_to_location(
    v: npt.NDArray[np.float64], loc: tuple[float, float]
) -> npt.NDArray[np.float64]:
    """
    Calculate the distance (in meters) between a passed lat/long location and each
    element in a vector of lat/long coordinates using the Haversine formula.

    :param v: A numpy array with shape (X, 2) with latitude, longitude coordinates
    :param loc: Tuple of latitude and longitude values caracterizing a location
    :return: Vector of distances between each point in the passed array and the passed
        location. Distance is returned in meters
    """
    _, rows = v.shape
    if rows != 2:
        raise RuntimeError("Pass an array with a (X, 2) shape")

    lats = v[:, 0:1]
    longs = v[:, 1:2]

    p_lat, p_long = loc

    p = pi / 180
    a = (
        0.5
        - np.cos((p_lat - lats) * pi) / 2
        + np.cos(lats * p) * np.cos(p_lat * p) * (1 - np.cos((p_long - longs) * p)) / 2
    )

    return 12742 * np.arcsin(np.sqrt(a)) * 1000


def get_point_distance(
    track: GPXTrack, segment_idx: None | int, latitude: float, longitude: float
) -> PointDistance:
    """
    Calculates the distance to the nearest point on a GPX track.

    :param track: The GPX track to analyze.
    :param segment_idx: The index of the segment to analyze. If None, all segments are
        analyzed.
    :param latitude: The latitude of the point to compare against.
    :param longitude: The longitude of the point to compare against.
    :raises TrackAnalysisError: If the nearest point could not be determined.

    :return: PointDistance: The calculated distance to the nearest point on the track.
    """
    points: list[tuple[float, float]] = []
    segment_point_idx_map: dict[int, tuple[int, int]] = {}
    if segment_idx is None:
        for i_segment, segment in enumerate(track.segments):
            first_idx = len(points)
            for point in segment.points:
                points.append((point.latitude, point.longitude))
            last_idx = len(points) - 1

            segment_point_idx_map[i_segment] = (first_idx, last_idx)
    else:
        segment = track.segments[segment_idx]
        for point in segment.points:
            points.append((point.latitude, point.longitude))

        segment_point_idx_map[segment_idx] = (0, len(points) - 1)

    distances = get_distances(np.array(points), np.array([[latitude, longitude]]))

    _min_idx = int(distances.argmin())
    min_distance = float(distances.min())
    _min_point = None
    _min_segment = -1
    _min_idx_in_segment = -1
    for i_seg, (i_min, i_max) in segment_point_idx_map.items():
        if i_min <= _min_idx <= i_max:
            _min_idx_in_segment = _min_idx - i_min
            _min_point = track.segments[i_seg].points[_min_idx_in_segment]
            _min_segment = i_seg
    if _min_point is None:
        raise TrackAnalysisError("Point could not be determined")

    return PointDistance(
        point=_min_point,
        distance=min_distance,
        point_idx_abs=_min_idx,
        segment_idx=_min_segment,
        segment_point_idx=_min_idx_in_segment,
    )


def get_points_inside_bounds(
    segment: GPXTrackSegment,
    bounds_min_latitude: float,
    bounds_min_longitude: float,
    bounds_max_latitude: float,
    bounds_max_longitude: float,
) -> list[tuple[int, bool]]:
    """
    Get a list of tuples representing points inside or outside a specified geographical
    bounds.

    :param segment: GPXTrackSegment to analyze.
    :param bounds_min_latitude: Minimum latitude of the geographical bounds.
    :param bounds_min_longitude: Minimum longitude of the geographical bounds.
    :param bounds_max_latitude: Maximum latitude of the geographical bounds.
    :param bounds_max_longitude: Maximum longitude of the geographical bounds.

    :return: List of tuples containing index and a boolean indicating whether the point
        is inside the bounds.
    """
    ret_list = []
    for idx, point in enumerate(segment.points):
        inside_bounds = (
            bounds_min_latitude <= point.latitude <= bounds_max_latitude
        ) and (bounds_min_longitude <= point.longitude <= bounds_max_longitude)
        ret_list.append((idx, inside_bounds))

    return ret_list


def split_segment_by_id(
    segment: GPXTrackSegment, index_ranges: list[tuple[int, int]]
) -> list[GPXTrackSegment]:
    """
    Split a GPXTrackSegment into multiple segments based on the provided index ranges.

    :param segment: GPXTrackSegment to be split.
    :param index_ranges: List of tuples representing index ranges for splitting the
        segment.

    :return: List of GPXTrackSegments resulting from the split.
    """
    ret_segments = []

    indv_idx: list[int] = []
    range_classifiers = []
    for range_ in index_ranges:
        indv_idx.extend(list(range_))
        range_classifiers.append(lambda i, le=range_[0], re=range_[1]: le <= i <= re)
        ret_segments.append(GPXTrackSegment())

    min_idx = min(indv_idx)
    max_idx = max(indv_idx)

    for idx, point in enumerate(segment.points):
        if idx < min_idx or idx > max_idx:
            continue

        for i_class, func in enumerate(range_classifiers):
            if func(idx):
                ret_segments[i_class].points.append(point)

    return ret_segments


def check_bounds(bounds: None | GPXBounds) -> None:
    """
    Check if the provided GPXBounds object is valid.

    :param bounds: GPXBounds object to be checked.
    :raises InvalidBoundsError: If the bounds object is None or has incomplete
        latitude/longitude values.
    """
    if bounds is None:
        raise InvalidBoundsError("Bounds %s are invalid", bounds)

    if (
        bounds.min_latitude is None
        or bounds.max_latitude is None
        or bounds.min_longitude is None
        or bounds.max_longitude is None
    ):
        raise InvalidBoundsError("Bounds %s are invalid", bounds)


def format_timedelta(td: timedelta) -> str:
    """
    Format a timedelta object as a string in HH:MM:SS format.

    :param td: Timedelta object to be formatted.

    :return: Formatted string representing the timedelta in HH:MM:SS format.
    """
    seconds = td.seconds
    hours = int(seconds / 3600)
    seconds -= hours * 3600
    minutes = int(seconds / 60)
    seconds -= minutes * 60

    if td.days > 0:
        hours += 24 * td.days

    return "{0:02d}:{1:02d}:{2:02d}".format(hours, minutes, seconds)


def fill_list(values: list[None | float]) -> list[float]:
    """Fills None values in a list with appropriate values. Leading (trailing) None
    values will be replace with first (last) real values. None in between real values
    will be linearly interpolated

    :param values: A list containing float values or None.

    :returns: A new list with all None values filled.

    Example:
        >>> filled_list = fill_list([None, 10.0, None, 20.0, None])
        >>> print(filled_list)  # Output: [10.0, 10.0, 15.0, 20.0, 20.0]
    """
    if set(values) == {None}:
        raise RuntimeError("At least on value must be not none")

    # Deal with leading None values
    if values[0] is None:
        logger.debug("Filling leading missing elevation values")
        first_real_elevation = next((x for x in values if x is not None), None)
        idx = 0
        while values[idx] is None:
            values[idx] = first_real_elevation
            idx += 1

    # Deal with trailing None values
    if values[-1] is None:
        idx = len(values) - 1
        trailing_none = []
        while values[idx] is None:
            trailing_none.append(idx)
            idx -= 1
        fill_value = values[idx]
        for _idx in trailing_none:
            values[_idx] = fill_value
    # Deal with values in between
    if None in values:
        consecutive_none_indices = []
        start_index = None
        for i, value in enumerate(values):
            if value is None:
                if start_index is None:
                    start_index = i
            else:
                if start_index is not None:
                    consecutive_none_indices.append((start_index, i))
                    start_index = None
        for idx_start, idx_end in consecutive_none_indices:
            new_values = interpolate_linear(
                values[idx_start - 1], values[idx_end], idx_end - idx_start + 1
            )
            for rp_idx, rp_value in enumerate(new_values[1:-1]):
                values[idx_start + rp_idx] = rp_value

    if None in values:
        raise RuntimeError
    return values  # type: ignore



class APIResponseError(Exception):
    pass


class APIHealthCheckFailedError(Exception):
    pass


class APIDataNotAvailableError(Exception):
    pass


class TrackInitializationError(Exception):
    pass


class DBTrackInitializationError(TrackInitializationError):
    pass


class TrackTransformationError(Exception):
    pass


class InvalidBoundsError(Exception):
    pass


class GPXPointExtensionError(Exception):
    pass


class VisualizationSetupError(Exception):
    pass


class TrackAnalysisError(Exception):
    pass


class VisualizationSetupWarning(Warning):
    pass


class GeoJsonWithoutGeometryError(Exception):
    pass


class UnsupportedGeoJsonTypeError(Exception):
    pass


class EmptyGeoJsonError(Exception):
    pass



import logging
from typing import Dict

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def hex_to_rgb(hex: str) -> tuple[int, int, int]:
    """
    Pass a hex color name (as string) and get the RGB value

    Source: https://medium.com/@BrendanArtley/matplotlib-color-gradients-21374910584b

    >> hex_to_RGB("#FFFFFF") -> [255,255,255]
    """
    return tuple([int(hex[i : i + 2], 16) for i in range(1, 6, 2)])  # type: ignore


def get_color_gradient(c1: str, c2: str, n: int) -> list[str]:
    """
    Create a color gradient between two passed colors with N steps.

    Source: https://medium.com/@BrendanArtley/matplotlib-color-gradients-21374910584b
    """
    assert n > 1
    c1_rgb = np.array(hex_to_rgb(c1)) / 255
    c2_rgb = np.array(hex_to_rgb(c2)) / 255
    mix_pcts = [x / (n - 1) for x in range(n)]
    rgb_colors = [((1 - mix) * c1_rgb + (mix * c2_rgb)) for mix in mix_pcts]
    return [
        ("#" + "".join([format(int(round(val * 255)), "02x") for val in item])).upper()
        for item in rgb_colors
    ]


def get_slope_colors(
    color_min: str,
    color_neutral: str,
    color_max: str,
    min_slope: int = -16,
    max_slope: int = 16,
) -> Dict[int, str]:
    """
    Generate a color gradient for the slope plots. The three passed colors are
    used for the MIN_SLOPE point, the 0 point and the MAX_SLOPE value respectively


    :param color_min: Color at the MIN_SLOPE value
    :param color_neutral: Color at 0
    :param color_max: Color at the MAX_SLOPE value
    :param min_slope: Minimum slope of the gradient, defaults to -16
    :param max_slope: Maximum slope of the gradient, defaults to 16
    :return: Dict mapping between slopes and colors
    """
    neg_points = list(range(min_slope, 1))
    pos_points = list(range(0, max_slope + 1))
    neg_colors = get_color_gradient(color_min, color_neutral, len(neg_points))
    pos_colors = get_color_gradient(color_neutral, color_max, len(pos_points))
    colors = {}
    colors.update({point: color for point, color in zip(neg_points, neg_colors)})
    colors.update({point: color for point, color in zip(pos_points, pos_colors)})
    return colors


def group_dataframe(
    data: pd.DataFrame, group_by: str, min_in_group: int
) -> list[pd.DataFrame]:
    """
    Group the DataFrame by a specified column, ensuring that each group contains at
    least a minimum number of rows. Except the first group, the last value from the
    previous group is prepended to the group to get consitent lines in a plot. Small
    groups (defined by min_in_group) will be merged in previous group. If this leads
    to multiple groups with same name, these will also be merged.

    :param data: The DataFrame to be grouped.
    :param group_by: The name of the column to group by.
    :param min_in_group: The minimum number of rows required in each group. Will be
        merged with previous group if number is not reached

    :return: A list of DataFrames, each containing a group of rows
        from the original DataFrame.
    """
    frames: list[pd.DataFrame] = []
    data["group"] = (data[group_by] != data[group_by].shift()).cumsum()
    group_by_color = f"{group_by}_colors".replace("zones", "zone")
    for _, group in data.groupby("group"):
        # First is always just added to the list
        if len(frames) == 0:
            frames.append(group)
            continue

        color = None

        if len(group) > min_in_group:
            group_name = group[group_by].unique()[0]
            if group_by_color in group:
                color = group[group_by_color].unique()[0]

            tral_prev_group = frames[len(frames) - 1].tail(1).copy()
            group = pd.concat([tral_prev_group, group])
            # Make sure the last entry from previous group hast same name
            group[group_by] = group_name
            if color is not None:
                group[group_by_color] = color
        else:
            group_name = frames[len(frames) - 1][group_by].unique()[0]
            if group_by_color in group:
                color = frames[len(frames) - 1][group_by_color].unique()[0]

            frames[len(frames) - 1] = pd.concat([frames[len(frames) - 1], group])
            frames[len(frames) - 1][group_by] = group_name
            if color is not None:
                frames[len(frames) - 1][group_by_color] = color
            continue

        name = group[group_by].unique()[0]
        prev_name = frames[len(frames) - 1][group_by].unique()[0]
        if name != prev_name:
            frames.append(group)
        else:
            # THis only happends with merged groups in between so groups with same name
            # so we need to drop the duplicates
            frames[len(frames) - 1] = pd.concat(
                [frames[len(frames) - 1], group]
            ).drop_duplicates()

    # Cleanup
    data.drop("group", axis=1, inplace=True)  # noqa: PD002
    return frames



ENRICH_UNITS: dict[str, str] = {
    "elevation": "m",
    "speed": "km/h",
    "heartrate": "bpm",
    "cadence": "rpm",
    "power": "W",
    "distance": "km",
    "temperature": "°C",
}

# Based on plotly.express.colors.sequential.Plotly3
DEFAULT_COLOR_GRADIENT: tuple[str, str] = ("#0508b8", "#fec3fe")

COLOR_GRADIENTS: dict[str, tuple[str, str]] = {
    "heartrate": ("#636EFA", "#EF553B"),
}

DEFAULT_BAR_COLORS: tuple[str, str] = ("#7570b3", "#8da0cb")



import logging
from collections import deque
from functools import lru_cache

import numpy as np
import numpy.typing as npt
from gpxpy.gpx import GPXTrack, GPXTrackSegment

from geo_track_analyzer.model import Position2D, SegmentOverlap
from geo_track_analyzer.utils.base import (
    check_bounds,
    crop_segment_to_bounds,
    distance,
    get_latitude_at_distance,
    get_longitude_at_distance,
    get_point_distance,
    get_points_inside_bounds,
    split_segment_by_id,
)

logger = logging.getLogger(__name__)


def check_segment_bound_overlap(
    reference_segments: GPXTrackSegment, segments: list[GPXTrackSegment]
) -> list[bool]:
    """Check if segments are within the bounds of a reference segment

    :param reference_segments: Segment defining the bounds inside which the other
        segments should be contained
    :param segments: Segments to be checked

    :return: List of bools specifying if the segments are inside the refence bounds
    """
    reference_bounds = reference_segments.get_bounds()

    check_bounds(reference_bounds)

    res = []

    for segment in segments:
        bounds = segment.get_bounds()

        check_bounds(bounds)

        res.append(
            bounds.min_latitude < reference_bounds.max_latitude  # type: ignore
            and bounds.min_longitude < reference_bounds.max_latitude  # type: ignore
            and bounds.max_latitude > reference_bounds.min_latitude  # type: ignore
            and bounds.max_longitude > reference_bounds.min_longitude  # type: ignore
        )

    return res


@lru_cache(100)
def derive_plate_bins(
    gird_width: float,
    bounds_min_latitude: float,
    bounds_min_longitude: float,
    bounds_max_latitude: float,
    bounds_max_longitude: float,
) -> tuple[list[tuple[float, float]], list[tuple[float, float]]]:
    """
    Derive the lat/long bins based on the min/max lat/long values and the target
    bin width.

    :param gird_width:  Lengths (in m) each bin will have in latitude and longitude
                        direction
    :param bounds_min_longitude: Minimum longitude of the grid
    :param bounds_max_latitude: Maximum latitude of the gtid. Bins may end with larger
                                values than passed here dependeing on the grid width
    :param bounds_max_longitude: Maximum longitude of the grid. Bins may end with
                                 larger values than passed here dependeing on the
                                 grid width

    :return: tuple with lists let/long values for the bin is latitude and longitude
             direction.
    """
    # Find the total distance in latitude and longitude directrons to find the number of
    # bins that need be generated bas ed on the pass gridwidth
    distance_latitude_total = gird_width + distance(
        Position2D(latitude=bounds_min_latitude, longitude=bounds_min_longitude),
        Position2D(latitude=bounds_max_latitude, longitude=bounds_min_longitude),
    )
    distance_longitude_total = gird_width + distance(
        Position2D(latitude=bounds_min_latitude, longitude=bounds_min_longitude),
        Position2D(latitude=bounds_min_latitude, longitude=bounds_max_longitude),
    )

    n_bins_latitude = int(round(distance_latitude_total / gird_width))
    n_bins_longitude = int(round(distance_longitude_total / gird_width))

    # The point at the lower bounds should be in the middle of the first bin. So
    # the edge of these bends need to be half the grid width from the original
    # lower left bound
    lower_edge_latitude = get_latitude_at_distance(
        Position2D(latitude=bounds_min_latitude, longitude=bounds_min_longitude),
        gird_width / 2,
        False,
    )
    lower_edge_longitude = get_longitude_at_distance(
        Position2D(latitude=bounds_min_latitude, longitude=bounds_min_longitude),
        gird_width / 2,
        False,
    )

    # Generate the bin edges by starting from the lower left edge and adding new
    # points with distance gid_width
    bins_latitude = [(lower_edge_latitude, lower_edge_longitude)]
    for _ in range(n_bins_latitude):
        _this_lat, _this_lon = bins_latitude[-1]
        bins_latitude.append(
            (
                get_latitude_at_distance(
                    Position2D(latitude=_this_lat, longitude=_this_lon),
                    gird_width,
                    True,
                ),
                lower_edge_longitude,
            )
        )

    bins_longitude = [(lower_edge_latitude, lower_edge_longitude)]
    for _ in range(n_bins_longitude):
        _this_lat, _this_lon = bins_longitude[-1]
        bins_longitude.append(
            (
                lower_edge_latitude,
                get_longitude_at_distance(
                    Position2D(latitude=_this_lat, longitude=_this_lon),
                    gird_width,
                    True,
                ),
            )
        )

    logger.debug(
        "Derived %s bins in latitude direction and %s in longitude direction",
        len(bins_latitude),
        len(bins_longitude),
    )
    logger.debug("  latitude direction: %s to %s", bins_latitude[0], bins_latitude[-1])
    logger.debug(
        "  longitude direction: %s to %s", bins_longitude[0], bins_longitude[-1]
    )

    return (bins_latitude, bins_longitude)


def convert_segment_to_plate(
    segment: GPXTrackSegment,
    gird_width: float,
    bounds_min_latitude: float,
    bounds_min_longitude: float,
    bounds_max_latitude: float,
    bounds_max_longitude: float,
    normalize: bool = False,
    max_queue_normalize: int = 5,
) -> np.ndarray:
    """
    Takes a GPXSegement and fills bins of a 2D array (called plate) with the passed
    bin width. Bins will start at the min latitude and longited values.

    :param segment: The GPXPoints of the Segment will be filled into the plate
    :param gird_width: Width (in meters) of the grid
    :param bounds_min_latitude: Minimum latitude of the grid
    :param bounds_min_longitude: Minimum longitude of the grid
    :param bounds_max_latitude: Maximum latitude of the gtid. Bins may end with larger
                                values than passed here dependeing on the grid width
    :param bounds_max_longitude: Maximum longitude of the grid. Bins may end with larger
                                values than passed here dependeing on the grid width
    :param normalize: If True, successive points (defined by the max_queue_normalize)
                      will not change the values in a bin. This means that each bin
                      values should have the value 1 except there is overlap with
                      points later in the track. To decide this the previous
                      max_queue_normalize points will be considered. So this value
                      dependes on the chosen gridwidth.
    :param max_queue_normalize: Number of previous bins considered when normalize is
                                set to true.

    :return: 2DArray representing the plate.
    """
    bins_latitude, bins_longitude = derive_plate_bins(
        gird_width,
        bounds_min_latitude,
        bounds_min_longitude,
        bounds_max_latitude,
        bounds_max_longitude,
    )

    _lat_bins = np.array([b[0] for b in bins_latitude])
    _long_bins = np.array([b[1] for b in bins_longitude])

    lats, longs = [], []
    for point in segment.points:
        lats.append(point.latitude)
        longs.append(point.longitude)

    # np.digitize starts with 1. We want 0 as first bin
    segment_lat_bins = np.digitize(lats, _lat_bins) - 1
    segment_long_bins = np.digitize(longs, _long_bins) - 1

    plate = np.zeros(shape=(len(bins_latitude), len(bins_longitude)))

    prev_bins = deque(maxlen=max_queue_normalize)  # type: ignore

    for lat, long in zip(segment_lat_bins, segment_long_bins):
        if normalize:
            if any((lat, long) == prev_bin for prev_bin in prev_bins):
                continue
            prev_bins.append((lat, long))
            plate[lat, long] += 1
        else:
            plate[lat, long] += 1

    return np.flip(plate, axis=0)


def _extract_ranges(
    base_points_in_bounds: list[tuple[int, bool]], allow_points_outside_bounds: int
) -> list[tuple[int, int]]:
    """Extract point ranges from a list of indices and boolan flags"""
    id_ranges_in_bounds: list[tuple[int, int]] = []
    found_range = False
    in_bound_range_start = -1
    points_since_last_range = 0
    idx = 0
    for idx, in_bounds in base_points_in_bounds:
        if in_bounds and not found_range:
            # Start of in_bound_points
            found_range = True
            if (
                points_since_last_range <= allow_points_outside_bounds
                and id_ranges_in_bounds
            ):
                prev_range_start, _ = id_ranges_in_bounds.pop()
                in_bound_range_start = prev_range_start
            else:
                in_bound_range_start = idx
        if not in_bounds:
            if found_range:
                # End of in_bound_points
                found_range = False
                id_ranges_in_bounds.append((in_bound_range_start, idx - 1))
                points_since_last_range = 0
            points_since_last_range += 1
    if found_range:
        id_ranges_in_bounds.append((in_bound_range_start, idx))

    return id_ranges_in_bounds


def get_segment_overlap(
    base_segment: GPXTrackSegment,
    match_segment: GPXTrackSegment,
    grid_width: float,
    max_queue_normalize: int = 5,
    allow_points_outside_bounds: int = 5,
    overlap_threshold: float = 0.75,
) -> list[SegmentOverlap]:
    """Compare the tracks of two segements and caclulate the overlap.

    :param base_segment: Base segement in which the match segment should be found
    :param match_segment: Other segmeent that should be found in the base segement
    :param grid_width: Width (in meters) of the grid that will be filled to estimate
                       the overalp.
    :param max_queue_normalize: Minimum number of successive points in the segment
                                between two points falling into same plate bin.
    :param allow_points_outside_bounds: Number of points between sub segments allowed
                                        for merging the segments.
    :param overlap_threshold: Minimum overlap required to return the overlap data.

    :return: list of SegmentOverlap objects.
    """
    bounds_match = match_segment.get_bounds()

    check_bounds(bounds_match)

    cropped_base_segment = crop_segment_to_bounds(
        base_segment,
        bounds_match.min_latitude,  # type: ignore
        bounds_match.min_longitude,  # type: ignore
        bounds_match.max_latitude,  # type: ignore
        bounds_match.max_longitude,  # type: ignore
    )

    plate_base = convert_segment_to_plate(
        cropped_base_segment,
        grid_width,
        bounds_match.min_latitude,  # type: ignore
        bounds_match.min_longitude,  # type: ignore
        bounds_match.max_latitude,  # type: ignore
        bounds_match.max_longitude,  # type: ignore
        True,
        max_queue_normalize,
    )

    plate_match = convert_segment_to_plate(
        match_segment,
        grid_width,
        bounds_match.min_latitude,  # type: ignore
        bounds_match.min_longitude,  # type: ignore
        bounds_match.max_latitude,  # type: ignore
        bounds_match.max_longitude,  # type: ignore
        True,
        max_queue_normalize,
    )

    # Check if the match segment appears muzltiple times in the base segemnt
    if plate_base.max() > 1:
        logger.debug(
            "Multiple occurances of points within match bounds in base segment"
        )
        base_points_in_bounds = get_points_inside_bounds(
            base_segment,
            bounds_match.min_latitude,  # type: ignore
            bounds_match.min_longitude,  # type: ignore
            bounds_match.max_latitude,  # type: ignore
            bounds_match.max_longitude,  # type: ignore
        )

        id_ranges_in_bounds = _extract_ranges(
            base_points_in_bounds, allow_points_outside_bounds
        )

        sub_segments = split_segment_by_id(base_segment, id_ranges_in_bounds)
        sub_segment_overlaps = []
        for i_sub_segment, (sub_segment, id_sub_segment) in enumerate(
            zip(sub_segments, id_ranges_in_bounds)
        ):
            logger.debug(
                "Processing overlap in sub-plate %s/%s",
                i_sub_segment + 1,
                len(sub_segments),
            )
            sub_plate = convert_segment_to_plate(
                sub_segment,
                grid_width,
                bounds_match.min_latitude,  # type: ignore
                bounds_match.min_longitude,  # type: ignore
                bounds_match.max_latitude,  # type: ignore
                bounds_match.max_longitude,  # type: ignore
                True,
                max_queue_normalize,
            )
            sub_segment_overlap = _calc_plate_overlap(
                base_segment=sub_segment,
                plate_base=sub_plate,
                match_segment=match_segment,
                plate_match=plate_match,
            )

            subseg_start, _ = id_sub_segment
            sub_segment_overlap.start_idx += subseg_start
            sub_segment_overlap.end_idx += subseg_start

            if sub_segment_overlap.overlap >= overlap_threshold:
                sub_segment_overlaps.append(sub_segment_overlap)

        return sorted(sub_segment_overlaps, key=lambda x: x.overlap, reverse=True)
    else:
        logger.debug("Processing overlap in plate")
        segment_overlap = _calc_plate_overlap(
            base_segment=base_segment,
            plate_base=plate_base,
            match_segment=match_segment,
            plate_match=plate_match,
        )
        if segment_overlap.overlap >= overlap_threshold:
            return [segment_overlap]
        else:
            return []


def _calc_plate_overlap(
    base_segment: GPXTrackSegment,
    plate_base: npt.NDArray,
    match_segment: GPXTrackSegment,
    plate_match: npt.NDArray,
) -> SegmentOverlap:
    overlap_plate = plate_base + plate_match

    overlap_plate_ = np.digitize(overlap_plate, np.array([0, 2, 3])) - 1

    overlapping_bins = np.sum(overlap_plate_)
    match_bins = np.sum(plate_match)

    logger.debug(
        "%s overlapping bins and %s bins in match segment", overlapping_bins, match_bins
    )

    overlap = overlapping_bins / match_bins

    logger.debug("Overlap: %s", overlap)

    # Determine if the direction in the base segmeent matched the direction
    # in the match segement
    first_point_match, last_point_match = (
        match_segment.points[0],
        match_segment.points[-1],
    )

    base_track = GPXTrack()
    base_track.segments = [base_segment]

    first_point_distance_info = get_point_distance(
        base_track, 0, first_point_match.latitude, first_point_match.longitude
    )
    first_point_base, first_idx = (
        first_point_distance_info.point,
        first_point_distance_info.segment_point_idx,
    )

    last_point_distance_info = get_point_distance(
        base_track, 0, last_point_match.latitude, last_point_match.longitude
    )

    last_point_base, last_idx = (
        last_point_distance_info.point,
        last_point_distance_info.segment_point_idx,
    )

    if last_idx > first_idx:
        logger.debug("Match direction: Same")
        inverse = False
        start_point, start_idx = first_point_base, first_idx
        end_point, end_idx = last_point_base, last_idx
    else:
        logger.debug("Match direction: Iverse")
        inverse = True
        end_point, end_idx = first_point_base, first_idx
        start_point, start_idx = last_point_base, last_idx

    return SegmentOverlap(
        overlap=overlap,
        inverse=inverse,
        plate=overlap_plate,
        start_point=start_point,
        start_idx=start_idx,
        end_point=end_point,
        end_idx=end_idx,
    )



import click
from requests.exceptions import MissingSchema
from rich.console import Console

from geo_track_analyzer.enhancer import EnhancerType, get_enhancer
from geo_track_analyzer.exceptions import (
    APIDataNotAvailableError,
    APIHealthCheckFailedError,
)
from geo_track_analyzer.track import GPXFileTrack


def convert_kwargs(raw_kwargs: tuple[str, ...]) -> dict[str, str | bool]:
    converted_kwargs: dict[str, str | bool] = {}
    for kwarg in raw_kwargs:
        if "=" not in kwarg:
            raise click.BadParameter(
                "All values in RAW_ENHANCER_INIT_ARGS need to contain =. "
                f"{kwarg} does not fulfill that requirement"
            )
        split_kwarg = kwarg.split("=")
        if len(split_kwarg) != 2:
            raise click.BadParameter(
                "All values in RAW_ENHANCER_INIT_ARGS need to be a value pair. "
                f"{kwarg} does not fulfill that requirement"
            )
        key, value = split_kwarg
        if value.lower() == "false":
            converted_kwargs[key] = False
        elif value.lower() == "true":
            converted_kwargs[key] = True
        else:
            converted_kwargs[key] = value

    return converted_kwargs


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "--enhancer",
    type=click.Choice(list(EnhancerType), case_sensitive=False),
    help="Specify the enhancer type to be used to make the requests.",
    required=True,
)
@click.option(
    "--url",
    help="URL of the API. Should be the full url including port if necessary. "
    "Example: http://localhost:8080/",
    required=True,
)
@click.option(
    "--postfix",
    default="enhanced_elevation",
    help="String that will be appended to the output file.",
    show_default=True,
)
@click.option(
    "-v", "--verbose", count=True, help="Set the verbosity level of the script."
)
@click.argument("raw_enhancer_init_args", nargs=-1, type=click.UNPROCESSED)
def main(
    filename: str,
    enhancer: EnhancerType,
    url: str,
    postfix: str,
    raw_enhancer_init_args: tuple[str, ...],
    verbose: int,
) -> None:
    """
    Enhance elevation data in a gpx file FILENAME using the Enahncer APIs
    provided in this package. Additional keyword-arguments for the Enhancers
    can be passed as key-value-pairs with = sizes. E.g. dataset=eudem25m for
    the OpenTopoElevationEnhancer.
    """
    console = Console()
    if not filename.endswith(".gpx"):
        raise click.UsageError("Only .gpx files are supported")
    init_kwargs = convert_kwargs(raw_enhancer_init_args)
    console.print(f"Enhancing [red]elevation[/red] for {filename}")

    try:
        track = GPXFileTrack(filename)
    except:  # noqa: E722
        console.print(":x: Track could not be loaded")
        exit(1)
    else:
        console.print(":white_check_mark: Track sucessfully loaded")

    enhancer_type = get_enhancer(enhancer)
    console.print(":white_check_mark: Enhancer selected")

    try:
        this_enhancer = enhancer_type(
            url=url,
            **init_kwargs,
        )
    except TypeError as e:
        console.print(":x: Enhancer could not be initialized. Check the passed kwargs")
        if verbose <= 1:
            console.print(f":exclamation_mark: {e}")
        exit(1)
    except MissingSchema as e:
        console.print(":x: Could not reach API")
        if verbose <= 1:
            console.print(f":exclamation_mark: {e}")
        exit(1)
    except APIHealthCheckFailedError as e:
        console.print(":x: Enhancer API Health-Check failed")
        if verbose <= 1:
            console.print(f":exclamation_mark: {e}")
        exit(1)
    except APIDataNotAvailableError as e:
        console.print(":x: Invalud dataset set. Check the passed kwargs")
        if verbose <= 1:
            console.print(f":exclamation_mark: {e}")
        exit(1)
    else:
        console.print(":white_check_mark: Enhancer initialized")

    with console.status("Running elevation enhancement"):
        this_enhancer.enhance_track(track.track, inplace=True)
    console.print(":white_check_mark: Enhancement done")

    new_file_name = filename.replace(".gpx", f"_{postfix}.gpx")
    with open(new_file_name, "w") as f:
        f.write(track.get_xml())

    console.print(f":white_check_mark: File saved to {new_file_name}")


if __name__ == "__main__":
    main()



import logging
import warnings
from enum import StrEnum, auto
from typing import Annotated, Callable

import pandas as pd
import plotly.graph_objects as go
from plotly.graph_objs import Figure
from plotly.subplots import make_subplots

from geo_track_analyzer.exceptions import (
    VisualizationSetupError,
    VisualizationSetupWarning,
)
from geo_track_analyzer.visualize.utils import group_dataframe

from .constants import ENRICH_UNITS

logger = logging.getLogger(__name__)


class PlotMetric(StrEnum):
    ELEVATION = auto()
    HEARTRATE = auto()
    POWER = auto()
    CADENCE = auto()
    SPEED = auto()
    TEMPERATURE = auto()


class PlotBase(StrEnum):
    DISTANCE = "cum_distance_moving"
    DURATION = "cum_time_moving"


METRICS_WITH_ZONES = [PlotMetric.CADENCE, PlotMetric.HEARTRATE, PlotMetric.POWER]


def _add_scatter_cont(
    fig: Figure,
    x: pd.Series,
    y: pd.Series,
    y_range: None | tuple[float, float],
    title: str,
    mode: str,
    fill: None | str = None,
    color: None | str = None,
    legend_group: None | str = None,
    show_legend: bool = False,
    y_title: None | str = None,
    is_secondary: bool = False,
) -> None:
    scatter_kwargs = dict(
        x=x,
        y=y,
        mode=mode,
        name=title,
        legendgroup=legend_group,
        fill=fill,
        showlegend=show_legend,
    )
    if color is not None:
        scatter_kwargs["line_color"] = color
    fig.add_trace(
        go.Scatter(**scatter_kwargs),
        secondary_y=is_secondary,
    )
    fig.update_yaxes(
        title_text=title if y_title is None else y_title,
        secondary_y=is_secondary,
        range=y_range,
    )


def _add_scatter_disonct(
    fig: Figure,
    data: pd.DataFrame,
    metric: str,
    y_converter: Callable[[pd.Series], pd.Series],
    y_range_max_factor: float,
    title: str,
    mode: str,
    fill: None | str = None,
    min_zone_size: float = 0.0025,
    is_secondary: bool = False,
    y_range_min_factor: float = 0,
) -> None:
    if f"{metric}_zones" not in data.columns:
        raise VisualizationSetupError("Zone data is not provided in passed dataframe")

    zone_data = group_dataframe(data, f"{metric}_zones", int(len(data) * min_zone_size))
    seen_group_names = []
    y_max = -99
    y_min = 99_999
    for data_ in zone_data[::-1]:
        group_name = data_[f"{metric}_zones"].iloc[-1]

        y_data = y_converter(data_[metric])
        y_max = max(y_max, y_data.max() * y_range_max_factor)
        y_min = min(y_min, y_data.min() * y_range_min_factor)
        _add_scatter_cont(
            fig=fig,
            x=data_.cum_distance_moving,
            y=y_data,
            y_range=None,
            title=group_name,
            mode=mode,
            fill=fill,
            color=data_[f"{metric}_zone_colors"].iloc[-1],
            legend_group=group_name,
            show_legend=group_name not in seen_group_names,
            y_title=title,
        )
        if group_name not in seen_group_names:
            seen_group_names.append(group_name)

    fig.update_yaxes(
        range=(y_min, y_max),
        secondary_y=is_secondary,
    )


def plot_metrics(
    data: pd.DataFrame,
    *,
    metrics: Annotated[list[PlotMetric], "Unique values"],
    base: PlotBase = PlotBase.DISTANCE,
    strict_data_selection: bool = False,
    height: int | None = 600,
    width: int | None = 1800,
    slider: bool = False,
    colors: list[str] | None = None,
    add_zones: bool = False,
) -> Figure:
    assert len(metrics) == len(set(metrics))

    if colors is None:
        _colors = [None for _ in metrics]
    else:
        if not len(colors) >= len(metrics):
            raise VisualizationSetupError(
                "Colors have been passed but at least "
                "the same number as metrics is required"
            )
        _colors = colors

    set_secondary = [False, True] if len(metrics) == 2 else [False for _ in metrics]
    if add_zones:
        zone_metrics = [m for m in metrics if m in METRICS_WITH_ZONES]
        if len(zone_metrics) == 0:
            add_zones = False
        elif len(zone_metrics) > 1:
            warnings.warn(
                VisualizationSetupWarning(
                    "Only one metric with zone is supported. Disabling zones"
                )
            )
            add_zones = False
        else:
            logger.debug("Adding zones for: %s", zone_metrics[0])
            # Make sure the zone metric is first
            _metrics = [zone_metrics[0]] + [m for m in metrics if m != zone_metrics[0]]
            _colors = [_colors[metrics.index(m)] for m in _metrics]
            metrics = _metrics
            set_secondary = [False] + [True for i in range(len(metrics) - 1)]

    mask = data.moving
    if strict_data_selection:
        mask = mask & data.in_speed_percentile

    data_for_plot: pd.DataFrame = data[mask].copy()  # type: ignore

    fig = make_subplots(specs=[[{"secondary_y": True}]])

    if base == PlotBase.DISTANCE:
        x_data = data_for_plot[base]
    else:
        x_data = pd.to_timedelta(data_for_plot[base], unit="s") + pd.Timestamp(
            "1970/01/01"
        )

    y_max = -99
    y_min = 99_999
    for metric, secondary, color in zip(metrics, set_secondary, _colors):
        if data_for_plot[metric].isna().all():
            if len(metrics) == 1:
                raise VisualizationSetupError(f"Cannot plot {metric}. Data missing")
            warnings.warn(
                VisualizationSetupWarning(f"Cannot plot {metric}. Data missing")
            )
            continue
        mode = "lines"
        y_converter: Callable[[pd.Series], pd.Series] = lambda s: s.fillna(0).astype(
            int
        )
        if metric == PlotMetric.CADENCE:
            mode = "markers"
        elif metric == PlotMetric.SPEED:
            y_converter = lambda s: s * 3.6

        y_data = y_converter(data_for_plot[metric])

        if add_zones and metric in METRICS_WITH_ZONES:
            fill = None
            if metric in [PlotMetric.HEARTRATE, PlotMetric.POWER]:
                fill = "tozeroy"

            _add_scatter_disonct(
                fig=fig,
                data=data_for_plot,
                metric=metric,
                mode=mode,
                fill=fill,
                y_converter=y_converter,
                y_range_max_factor=1.2,
                y_range_min_factor=0.6,
                title=f"{metric.capitalize()} [{ENRICH_UNITS[metric]}]",
                is_secondary=secondary,
            )
        else:
            y_max = max(y_max, y_data.max() * 1.2)
            dmin = y_data.min()
            y_min = min(y_min, dmin * 0.6 if dmin > 0 else dmin * 1.5)
            _add_scatter_cont(
                fig=fig,
                x=x_data,
                y=y_data,
                mode=mode,
                color=color,
                title=f"{metric.capitalize()} [{ENRICH_UNITS[metric]}]",
                y_range=None,
                is_secondary=secondary,
                show_legend=len(metrics) > 2,
            )
    if slider:
        fig.update_layout(
            xaxis=dict(
                rangeslider=dict(visible=True),
            )
        )

    if base == PlotBase.DISTANCE:
        fig.update_xaxes(title_text="Distance [m]")
    else:
        fig.update_layout(
            xaxis=dict(
                title="Duration [HH:MM:SS]",
                tickformat="%H:%M:%S",
            )
        )
    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    if len(metrics) == 1:
        metric = metrics[0]
        fig.update_yaxes(title_text=f"{metric.capitalize()} [{ENRICH_UNITS[metric]}]")
    elif len(metrics) > 2:
        fig.update_yaxes(title_text=None)
        fig.update_yaxes(range=(y_min, y_max), secondary_y=add_zones)
    return fig



from __future__ import annotations

import json
import logging
import warnings
from abc import ABC, abstractmethod
from copy import copy
from datetime import datetime
from io import StringIO
from itertools import pairwise
from typing import TYPE_CHECKING, Dict, Literal, Sequence, TypeVar, final

import gpxpy
import numpy as np
import pandas as pd
from fitparse import DataMessage, FitFile, StandardUnitsDataProcessor
from gpxpy.gpx import GPX, GPXTrack, GPXTrackSegment

if TYPE_CHECKING:
    from plotly.graph_objs.graph_objs import Figure

from geo_track_analyzer.compare import get_segment_overlap
from geo_track_analyzer.exceptions import (
    TrackInitializationError,
    TrackTransformationError,
    VisualizationSetupError,
)
from geo_track_analyzer.model import (
    PointDistance,
    Position3D,
    SegmentOverview,
    SegmentOverviewMetric,
    Zones,
)
from geo_track_analyzer.processing import (
    get_processed_segment_data,
    get_processed_track_data,
)
from geo_track_analyzer.utils.base import (
    calc_elevation_metrics,
    fill_list,
    get_point_distance,
    interpolate_segment,
)
from geo_track_analyzer.utils.internal import (
    BackFillExtensionDict,
    _points_eq,
    get_extended_track_point,
    get_extensions_in_points,
)
from geo_track_analyzer.visualize import (
    plot_metrics,
    plot_segment_box_summary,
    plot_segment_summary,
    plot_segment_zones,
    plot_segments_on_map,
    plot_track_2d,
    plot_track_enriched_on_map,
    plot_track_line_on_map,
    plot_track_with_slope,
    plot_track_zones,
)

N = TypeVar("N", int, float)

logger = logging.getLogger(__name__)

process_data_tuple_type = tuple[float, float, float, float, pd.DataFrame]


class Track(ABC):
    """
    Abstract base container for geospacial Tracks that defines all methods common to
    all Track types.
    """

    def __init__(
        self,
        stopped_speed_threshold: float,
        max_speed_percentile: int,
        extensions: set[str] | None = None,
        require_data_extensions: set[str] | None = None,
        heartrate_zones: None | Zones = None,
        power_zones: None | Zones = None,
        cadence_zones: None | Zones = None,
    ) -> None:
        logger.debug(
            "Using threshold for stopped speed: %s km/h", stopped_speed_threshold
        )
        logger.debug("Using %s percentile to calculate overview", max_speed_percentile)

        self.stopped_speed_threshold = stopped_speed_threshold
        self.max_speed_percentile = max_speed_percentile

        self._processed_segment_data: dict[int, process_data_tuple_type] = {}
        self._processed_track_data: dict[str, tuple[int, process_data_tuple_type]] = {}

        self.session_data: Dict[str, str | int | float] = {}

        self.require_data_extensions = require_data_extensions
        self.extensions: set[str] = set() if extensions is None else extensions
        self.heartrate_zones = heartrate_zones
        self.power_zones = power_zones
        self.cadence_zones = cadence_zones

    @property
    @abstractmethod
    def track(self) -> GPXTrack: ...

    @property
    def n_segments(self) -> int:
        return len(self.track.segments)

    def add_segmeent(self, segment: GPXTrackSegment) -> None:
        """Add a new segment ot the track

        :param segment: GPXTracksegment to be added
        """
        self.track.segments.append(segment)
        logger.info("Added segment with postition: %s", len(self.track.segments))

    def strip_segements(self) -> bool:
        """
        Strip all segments from the track. Duplicate points at the segmentment boardes
        will be dropped.
        """
        while len(self.track.segments) != 1:
            if not self.remove_segement(len(self.track.segments) - 1, "before"):
                return False

        return True

    def remove_segement(
        self, n_segment: int, merge: Literal["before", "after"] = "before"
    ) -> bool:
        """
        Remove a given segment from the track and merge it with the previous or next
        segment. Will return False, of the passed parameters lead to not possible
        operation.

        :param n_segment: Index of the segment the overview should be generated for,
        :param merge: Direction of the merge. Possible values of "before" and "after".

        :return: Boolean value reflecting if a segment was removed
        """
        assert merge in ["before", "after"]
        if n_segment == 0 and merge != "after":
            logger.error("First segement can only be merged with the after method")
            return False
        if merge == "after" and n_segment == len(self.track.segments) - 1:
            logger.error("Last segment can only be merged with the before method")
            return False
        try:
            self.track.segments[n_segment]
        except IndexError:
            logger.error(
                "Cannot remove segment %s. No valid key in segments", n_segment
            )
            return False

        if merge == "after":
            idx_end = None
            if _points_eq(
                self.track.segments[n_segment].points[-1],
                self.track.segments[n_segment + 1].points[0],
            ):
                idx_end = -1
            self.track.segments[n_segment + 1].points = (
                self.track.segments[n_segment].points[0:idx_end]
                + self.track.segments[n_segment + 1].points
            )
            self.track.segments.pop(n_segment)
            return True
        else:
            idx_start = 0
            if _points_eq(
                self.track.segments[n_segment].points[0],
                self.track.segments[n_segment - 1].points[-1],
            ):
                idx_start = 1
            self.track.segments[n_segment - 1].points = (
                self.track.segments[n_segment - 1].points
                + self.track.segments[n_segment].points[idx_start:]
            )
            self.track.segments.pop(n_segment)
            return True

    def get_xml(self, name: None | str = None, email: None | str = None) -> str:
        """Get track as .gpx file data

        :param name: Optional author name to be added to gpx file, defaults to None
        :param email: Optional auther e-mail address to be added to the gpx file,
            defaults to None

        :return: Content of a gpx file
        """
        gpx = GPX()

        gpx.tracks = [self.track]
        gpx.author_name = name
        gpx.author_email = email

        return gpx.to_xml()

    def _update_extensions(self) -> None:
        for segment in self.track.segments:
            self.extensions.update(get_extensions_in_points(segment.points))

    def get_track_overview(
        self, connect_segments: Literal["full", "forward"] = "forward"
    ) -> SegmentOverview:
        """
        Get overall metrics for the track. Equivalent to the sum of all segments

        :return: A SegmentOverview object containing the metrics
        """
        (
            track_time,
            track_distance,
            track_stopped_time,
            track_stopped_distance,
            track_data,
        ) = self._get_processed_track_data(connect_segments=connect_segments)

        track_max_speed = None
        track_avg_speed = None

        if (
            all(seg.has_times() for seg in self.track.segments)
            and not track_data.speed.isna().all()
        ):
            track_max_speed = track_data.speed[track_data.in_speed_percentile].max()
            track_avg_speed = track_data.speed[track_data.in_speed_percentile].mean()

        return self._create_segment_overview(
            time=track_time,
            distance=track_distance,
            stopped_time=track_stopped_time,
            stopped_distance=track_stopped_distance,
            max_speed=track_max_speed,
            avg_speed=track_avg_speed,
            data=track_data,  # type: ignore
        )

    def get_segment_overview(self, n_segment: int = 0) -> SegmentOverview:
        """
        Get overall metrics for a segment

        :param n_segment: Index of the segment the overview should be generated for,
            default to 0

        :returns: A SegmentOverview object containing the metrics moving time and
            distance, total time and distance, maximum and average speed and elevation
            and cummulated uphill, downholl elevation
        """
        (
            time,
            distance,
            stopped_time,
            stopped_distance,
            data,
        ) = self._get_processed_segment_data(n_segment)

        max_speed = None
        avg_speed = None

        if self.track.segments[n_segment].has_times():
            max_speed = data.speed[data.in_speed_percentile].max()
            avg_speed = data.speed[data.in_speed_percentile].mean()

        return self._create_segment_overview(
            time=time,
            distance=distance,
            stopped_time=stopped_time,
            stopped_distance=stopped_distance,
            max_speed=max_speed,
            avg_speed=avg_speed,
            data=data,
        )

    def _create_segment_overview(
        self,
        time: float,
        distance: float,
        stopped_time: float,
        stopped_distance: float,
        max_speed: None | float,
        avg_speed: None | float,
        data: pd.DataFrame,
    ) -> SegmentOverview:
        """Derive overview metrics for a segmeent"""
        total_time = time + stopped_time
        total_distance = distance + stopped_distance

        max_elevation = None
        min_elevation = None

        uphill = None
        downhill = None

        if (max_speed is None and avg_speed is not None) or (
            max_speed is not None and avg_speed is None
        ):
            raise RuntimeError("You need to passed max and avg speed or none")
        if not data.elevation.isna().all():
            max_elevation = data.elevation.max()
            min_elevation = data.elevation.min()
            position_3d = [
                Position3D(
                    latitude=rec["latitude"],
                    longitude=rec["longitude"],
                    elevation=rec["elevation"],
                )
                for rec in data.to_dict("records")
                if not np.isnan(rec["elevation"])
            ]
            elevation_metrics = calc_elevation_metrics(position_3d)

            uphill = elevation_metrics.uphill
            downhill = elevation_metrics.downhill
        overview_metrics = {}
        for metric in ["heartrate", "cadence", "power"]:
            overview_metrics[metric] = None
            if metric in data.columns and not data[metric].isna().all():
                overview_metrics[metric] = SegmentOverviewMetric[int](
                    max=int(round(data[metric].max(), 0)),
                    avg=int(round(data[metric].mean(), 0)),
                )

        return SegmentOverview(
            moving_time_seconds=time,
            total_time_seconds=total_time,
            moving_distance=distance,
            total_distance=total_distance,
            velocity=SegmentOverviewMetric[float](
                max=max_speed,
                avg=avg_speed,  # type: ignore
            )
            if max_speed is not None
            else None,
            max_elevation=max_elevation,
            min_elevation=min_elevation,
            uphill_elevation=uphill,
            downhill_elevation=downhill,
            cadence=overview_metrics["cadence"],
            power=overview_metrics["power"],
            heartrate=overview_metrics["heartrate"],
        )

    def get_closest_point(
        self, n_segment: None | int, latitude: float, longitude: float
    ) -> PointDistance:
        """
        Get closest point in a segment or track to the passed latitude and longitude
        corrdinate

        :param n_segment: Index of the segment. If None is passed the whole track is
            considered
        :param latitude: Latitude to check
        :param longitude: Longitude to check

        :return: Tuple containg the point as GPXTrackPoint, the distance from
            the passed coordinates and the index in the segment
        """
        return get_point_distance(self.track, n_segment, latitude, longitude)

    def _get_aggregated_pp_distance(self, agg: str, threshold: float) -> float:
        data = self.get_track_data()

        return data[data.distance >= threshold].distance.agg(agg)

    def _get_aggregated_pp_distance_in_segment(
        self, agg: str, n_segment: int, threshold: float
    ) -> float:
        data = self.get_segment_data(n_segment=n_segment)

        return data[data.distance >= threshold].distance.agg(agg)

    def get_avg_pp_distance(self, threshold: float = 10) -> float:
        """
        Get average distance between points in the track.

        :param threshold: Minimum distance between points required to  be used for the
            average, defaults to 10

        :return: Average distance
        """
        return self._get_aggregated_pp_distance("average", threshold)

    def get_avg_pp_distance_in_segment(
        self, n_segment: int = 0, threshold: float = 10
    ) -> float:
        """
        Get average distance between points in the segment with index n_segment.

        :param n_segment: Index of the segement to process, defaults to 0
        :param threshold: Minimum distance between points required to  be used for the
            average, defaults to 10

        :return: Average distance
        """
        return self._get_aggregated_pp_distance_in_segment(
            "average", n_segment, threshold
        )

    def get_max_pp_distance(self, threshold: float = 10) -> float:
        """
        Get maximum distance between points in the track.

        :param threshold: Minimum distance between points required to  be used for the
            maximum, defaults to 10

        :return: Maximum distance
        """
        return self._get_aggregated_pp_distance("max", threshold)

    def get_max_pp_distance_in_segment(
        self, n_segment: int = 0, threshold: float = 10
    ) -> float:
        """
        Get maximum distance between points in the segment with index n_segment.

        :param n_segment: Index of the segement to process, defaults to 0
        :param threshold: Minimum distance between points required to  be used for the
            maximum, defaults to 10

        :return: Maximum distance
        """
        return self._get_aggregated_pp_distance_in_segment("max", n_segment, threshold)

    def _get_processed_segment_data(
        self, n_segment: int = 0
    ) -> tuple[float, float, float, float, pd.DataFrame]:
        if n_segment not in self._processed_segment_data:
            (
                time,
                distance,
                stopped_time,
                stopped_distance,
                data,
            ) = get_processed_segment_data(
                self.track.segments[n_segment],
                self.stopped_speed_threshold,
                heartrate_zones=self.heartrate_zones,
                power_zones=self.power_zones,
                cadence_zones=self.cadence_zones,
                extensions=self.extensions,
                require_extensions=self.require_data_extensions,
            )

            if data.time.notna().any():
                data = self._apply_outlier_cleaning(data)

            self._processed_segment_data[n_segment] = (
                time,
                distance,
                stopped_time,
                stopped_distance,
                data,
            )

        return self._processed_segment_data[n_segment]

    def _get_processed_track_data(
        self, connect_segments: Literal["full", "forward"]
    ) -> process_data_tuple_type:
        if connect_segments in self._processed_track_data:
            segments_in_data, data = self._processed_track_data[connect_segments]
            if segments_in_data == self.n_segments:
                return data

        (
            time,
            distance,
            stopped_time,
            stopped_distance,
            processed_data,
        ) = get_processed_track_data(
            self.track,
            self.stopped_speed_threshold,
            connect_segments=connect_segments,
            heartrate_zones=self.heartrate_zones,
            power_zones=self.power_zones,
            cadence_zones=self.cadence_zones,
            extensions=self.extensions,
            require_extensions=self.require_data_extensions,
        )

        if processed_data.time.notna().any():
            processed_data = self._apply_outlier_cleaning(processed_data)

        return self._set_processed_track_data(
            (
                time,
                distance,
                stopped_time,
                stopped_distance,
                processed_data,
            ),
            connect_segments,
        )

    def _set_processed_track_data(
        self,
        data: process_data_tuple_type,
        connect_segments: Literal["full", "forward"],
    ) -> process_data_tuple_type:
        """Save processed data internally to reduce compute.
        Mainly separated for testing"""
        self._processed_track_data[connect_segments] = (self.n_segments, data)
        return data

    def get_segment_data(self, n_segment: int = 0) -> pd.DataFrame:
        """Get processed data for the segmeent with passed index as DataFrame

        :param n_segment: Index of the segement, defaults to 0

        :return: DataFrame with segmenet data
        """
        _, _, _, _, data = self._get_processed_segment_data(n_segment)

        return data

    def get_track_data(
        self, connect_segments: Literal["full", "forward"] = "forward"
    ) -> pd.DataFrame:
        """
        Get processed data for the track as DataFrame. Segment are indicated
        via the segment column.

        :return: DataFrame with track data
        """
        track_data: None | pd.DataFrame = None

        _, _, _, _, track_data = self._get_processed_track_data(
            connect_segments=connect_segments
        )

        return track_data

    def interpolate_points_in_segment(
        self,
        spacing: float,
        n_segment: int = 0,
        copy_extensions: Literal[
            "copy-forward", "meet-center", "linear"
        ] = "copy-forward",
    ) -> None:
        """
        Add additdion points to a segment by interpolating along the direct line
        between each point pair according to the passed spacing parameter. If present,
        elevation and time will be linearly interpolated. Extensions (Heartrate,
        Cadence, Power) will be interpolated according to value of copy_extensions.
        Optionas are:

        - copy the value from the start point of the interpolation (copy-forward)
        - Use value of start point for first half and last point for second half
          (meet-center)
        - Linear interpolation (linear)


        :param spacing: Minimum distance between points added by the interpolation
        :param n_segment: segment in the track to use, defaults to 0
        :param copy_extensions: How should the extenstion (if present) be defined in the
            interpolated points.
        """
        self.track.segments[n_segment] = interpolate_segment(
            self.track.segments[n_segment], spacing, copy_extensions=copy_extensions
        )

        # Reset saved processed data
        for key in self._processed_track_data:
            self._processed_track_data.pop(key)
        if n_segment in self._processed_segment_data:
            logger.debug(
                "Deleting saved processed segment data for segment %s", n_segment
            )
            self._processed_segment_data.pop(n_segment)

    def get_point_data_in_segmnet(
        self, n_segment: int = 0
    ) -> tuple[list[tuple[float, float]], None | list[float], None | list[datetime]]:
        """Get raw coordinates (latitude, longitude), times and elevations for the
        segement with the passed index.

        :param n_segment: Index of the segement, defaults to 0

        :return: tuple with coordinates (latitude, longitude), times and elevations
        """
        coords = []
        elevations = []
        times = []

        for point in self.track.segments[n_segment].points:
            coords.append((point.latitude, point.longitude))
            if point.elevation is not None:
                elevations.append(point.elevation)
            if point.time is not None:
                times.append(point.time)

        if not elevations:
            elevations = None  # type: ignore
        elif len(coords) != len(elevations):
            raise TrackTransformationError(
                "Elevation is not set for all points. This is not supported"
            )
        if not times:
            times = None  # type: ignore
        elif len(coords) != len(times):
            raise TrackTransformationError(
                "Elevation is not set for all points. This is not supported"
            )

        return coords, elevations, times

    def _apply_outlier_cleaning(self, data: pd.DataFrame) -> pd.DataFrame:
        speeds = data.speed[data.speed.notna()].to_list()
        if not speeds:
            logger.warning(
                "Trying to apply outlier cleaning to track w/o speed information"
            )
            return data
        speed_percentile = np.percentile(
            data.speed[data.speed.notna()].to_list(),
            self.max_speed_percentile,
        )

        data_ = data.copy()

        data_["in_speed_percentile"] = data_.apply(
            lambda c: c.speed <= speed_percentile, axis=1
        )

        return data_

    def find_overlap_with_segment(
        self,
        n_segment: int,
        match_track: Track,
        match_track_segment: int = 0,
        width: float = 50,
        overlap_threshold: float = 0.75,
        max_queue_normalize: int = 5,
        merge_subsegments: int = 5,
        extensions_interpolation: Literal[
            "copy-forward", "meet-center", "linear"
        ] = "copy-forward",
    ) -> Sequence[tuple[Track, float, bool]]:
        """Find overlap of a segment of the track with a segment in another track.

        :param n_segment: Segment in the track that sould be used as base for the
            comparison
        :param match_track: Track object containing the segment to be matched
        :param match_track_segment: Segment on the passed track that should be matched
            to the segment in this track, defaults to 0
        :param width: Width (in meters) of the grid that will be filled to estimate
            the overalp , defaults to 50
        :param overlap_threshold: Minimum overlap (as fracrtion) required to return the
            overlap data, defaults to 0.75
        :param max_queue_normalize: Minimum number of successive points in the segment
            between two points falling into same plate bin, defaults to 5
        :param merge_subsegments: Number of points between sub segments allowed
            for merging the segments, defaults to 5
        :param extensions_interpolation: How should the extenstion (if present) be
            defined in the interpolated points, defaults to copy-forward

        :return: Tuple containing a Track with the overlapping points, the overlap in
            percent, and the direction of the overlap
        """
        max_distance_self = self.get_max_pp_distance_in_segment(n_segment)

        segment_self = self.track.segments[n_segment]
        if max_distance_self > width:
            segment_self = interpolate_segment(
                segment_self, width / 2, copy_extensions=extensions_interpolation
            )

        max_distance_match = match_track.get_max_pp_distance_in_segment(
            match_track_segment
        )
        segment_match = match_track.track.segments[match_track_segment]
        if max_distance_match > width:
            segment_match = interpolate_segment(
                segment_match, width / 2, copy_extensions=extensions_interpolation
            )

        logger.info("Looking for overlapping segments")
        segment_overlaps = get_segment_overlap(
            segment_self,
            segment_match,
            width,
            max_queue_normalize,
            merge_subsegments,
            overlap_threshold,
        )

        matched_tracks: list[tuple[Track, float, bool]] = []
        for overlap in segment_overlaps:
            logger.info("Found: %s", overlap)
            matched_segment = GPXTrackSegment()
            # TODO: Might need to go up to overlap.end_idx + 1?
            matched_segment.points = self.track.segments[n_segment].points[
                overlap.start_idx : overlap.end_idx
            ]
            matched_tracks.append(
                (
                    SegmentTrack(
                        matched_segment,
                        stopped_speed_threshold=self.stopped_speed_threshold,
                        max_speed_percentile=self.max_speed_percentile,
                    ),
                    overlap.overlap,
                    overlap.inverse,
                )
            )
        return matched_tracks

    def plot(
        self,
        kind: Literal[
            "profile",
            "profile-slope",
            "map-line",
            "map-line-enhanced",
            "map-segments",
            "zone-summary",
            "segment-zone-summary",
            "segment-box",
            "segment-summary",
            "metrics",
        ],
        *,
        segment: None | int | list[int] = None,
        reduce_pp_intervals: None | int = None,
        use_distance_segments: None | float = None,
        **kwargs,
    ) -> Figure:
        """
        Visualize the full track or a single segment.

        :param kind: Kind of plot to be generated

            - profile: Elevation profile of the track. May be enhanced with additional
              information like Velocity, Heartrate, Cadence, and Power. Pass keyword
              args for [`plot_track_2d`][geo_track_analyzer.visualize.plot_track_2d]
            - profile-slope: Elevation profile with slopes between points. Use the
              reduce_pp_intervals argument to reduce the number of slope intervals.
              Pass keyword args for
              [`plot_track_with_slope`][geo_track_analyzer.visualize.plot_track_with_slope]
            - map-line: Visualize coordinates on the map. Pass keyword args for
              [`plot_track_line_on_map`][geo_track_analyzer.visualize.plot_track_line_on_map]
            - map-line-enhanced: Visualize coordinates on the map. Enhance with
              additional information like Elevation, Velocity, Heartrate, Cadence, and
              Power. Pass keyword args for [`plot_track_enriched_on_map`][geo_track_analyzer.visualize.plot_track_enriched_on_map]
            - map-segments: Visualize coordinates on the map split into segments.
              Pass keyword args for
              [`plot_segments_on_map`][geo_track_analyzer.visualize.plot_segments_on_map]
            - zone_summary : Visualize an aggregate (time, distance, speed) value for a
                metric (heartrate, power, cadence) with defined zones. Pass keyword args
                for [`plot_track_zones`][geo_track_analyzer.visualize.plot_track_zones],
                `aggregate` and `metric` are required.
            - segment_zone_summary : Same as "zone-summary" but split aggregate per
                segment [`plot_segment_zones`][geo_track_analyzer.visualize.plot_segment_zones]
            - segment_box : Box plot of a metric (heartrate, power, cadence, speed,
                elevation) per segment. Pass keyword args for [`plot_segments_on_map`][geo_track_analyzer.visualize.plot_segments_on_map]
                `metric` is required.
            - segment_summary : Visualize a aggregate (total_time, total_distance,
                avg_speed, max_speed) per segment. Pass keyword args for [`plot_segment_summary`][geo_track_analyzer.visualize.plot_segment_summary]
                `aggregate` is required.
            - metrics: Visualize the progression of of a metric (elevation, heartrate,
                power, cadence, power) over the course of a track. Can be plotted over distance
                and duration.
        :param segment: Select a specific segment, multiple segments or all segmenets,
            defaults to None
        :param reduce_pp_intervals: Optionally pass a distance in m which is used to
            reduce the points in a track, defaults to None
        :param use_distance_segments: Ignore all segments in data and split full track
            into segments with passed cummulated distance in meters. If passed, segment
            arg must be None. Defaults to None.
        :raises VisualizationSetupError: If the plot prequisites are not met

        :return: Figure (plotly)
        """
        from geo_track_analyzer.utils.track import generate_distance_segments

        if use_distance_segments is not None and segment is not None:
            raise VisualizationSetupError(
                f"use_distance_segments {use_distance_segments} cannot be passed with "
                f"segment {segment}."
            )

        valid_kinds = [
            "profile",
            "profile-slope",
            "map-line",
            "map-line-enhanced",
            "map-segments",
            "zone-summary",
            "segment-zone-summary",
            "segment-box",
            "segment-summary",
            "metrics",
        ]

        if "_" in kind and kind.replace("_", "-") in valid_kinds:
            warnings.warn(
                "Found %s but in versions >=2 only %s will be supported"
                % (kind, kind.replace("_", "-")),
                DeprecationWarning,
            )
            kind = kind.replace("_", "-")  # type: ignore

        require_elevation = ["profile", "profile-slope"]
        connect_segment_full = ["map-segments"]
        if kind not in valid_kinds:
            raise VisualizationSetupError(
                f"Kind {kind} is not valid. Pass on of {','.join(valid_kinds)}"
            )

        if kind in ["zone-summary", "segment-zone-summary"] and not all(
            key in kwargs for key in ["metric", "aggregate"]
        ):
            raise VisualizationSetupError(
                f"If {kind} is passed, **metric** and **aggregate** need to be passed"
            )
        if kind in ["segment-box"] and not all(key in kwargs for key in ["metric"]):
            raise VisualizationSetupError(
                f"If {kind} is passed, **metric** needs to be passed"
            )
        if kind in ["segment-summary"] and not all(
            key in kwargs for key in ["aggregate"]
        ):
            raise VisualizationSetupError(
                f"If {kind} is passed, **metric** needs to be passed"
            )

        require_extensions = (
            set()
            if self.require_data_extensions is None
            else copy(self.require_data_extensions)
        )
        require_extensions.update({"heartrate", "power", "cadence"})

        if segment is None:
            from geo_track_analyzer.utils.track import extract_track_data_for_plot

            data = extract_track_data_for_plot(
                track=self,
                kind=kind,
                require_elevation=require_elevation,
                intervals=reduce_pp_intervals,
                connect_segments="full" if kind in connect_segment_full else "forward",
                extensions=self.extensions,
            )
        elif isinstance(segment, int):
            from geo_track_analyzer.utils.track import extract_segment_data_for_plot

            data = extract_segment_data_for_plot(
                track=self,
                segment=segment,
                kind=kind,
                require_elevation=require_elevation,
                intervals=reduce_pp_intervals,
                extensions=self.extensions,
            )
        else:
            from geo_track_analyzer.utils.track import (
                extract_multiple_segment_data_for_plot,
            )

            data = extract_multiple_segment_data_for_plot(
                track=self,
                segments=segment,
                kind=kind,
                require_elevation=require_elevation,
                intervals=reduce_pp_intervals,
                extensions=self.extensions,
            )

        # Special case if all points are stopped. Since we still want
        # a plot, we interpret it as all moving. This is expected behaviour
        # for tracks that all have the same lat/lon coordinates
        if (data.moving == False).all():  # noqa: E712
            data.moving = True
            data.cum_time_moving = data.cum_time_stopped

        if use_distance_segments is not None:
            data = generate_distance_segments(data, use_distance_segments)

        fig: Figure
        if kind == "profile":
            fig = plot_track_2d(data=data, **kwargs)
        elif kind == "profile-slope":
            fig = plot_track_with_slope(data=data, **kwargs)
        elif kind == "map-line":
            fig = plot_track_line_on_map(data=data, **kwargs)
        elif kind == "map-line-enhanced":
            fig = plot_track_enriched_on_map(data=data, **kwargs)
        elif kind == "map-segments":
            fig = plot_segments_on_map(data=data, **kwargs)
        elif kind == "zone-summary":
            fig = plot_track_zones(data=data, **kwargs)
        elif kind == "segment-zone-summary":
            fig = plot_segment_zones(data=data, **kwargs)
        elif kind == "segment-summary":
            fig = plot_segment_summary(data=data, **kwargs)
        elif kind == "segment-box":
            fig = plot_segment_box_summary(data=data, **kwargs)
        elif kind == "metrics":
            fig = plot_metrics(data=data, **kwargs)

        return fig

    def split(
        self, coords: tuple[float, float], distance_threshold: float = 20
    ) -> None:
        """
        Split the track at the passed coordinates. The distance_threshold parameter
        defines the maximum distance between the passed coordingates and the closest
        point in the track.

        :param coords: Latitude, Longitude point at which the split should be made
        :param distance_threshold: Maximum distance between coords and closest point,
            defaults to 20

        :raises TrackTransformationError: If distance exceeds threshold
        """
        lat, long = coords
        point_distance = get_point_distance(
            self.track, None, latitude=lat, longitude=long
        )

        if point_distance.distance > distance_threshold:
            raise TrackTransformationError(
                f"Closes point in track has distance {point_distance.distance:.2f}m "
                "from passed coordingates"
            )
        # Split the segment. The closest point should be the first
        # point of the second segment
        pre_segment, post_segment = self.track.segments[
            point_distance.segment_idx
        ].split(point_distance.segment_point_idx - 1)

        self.track.segments[point_distance.segment_idx] = pre_segment
        self.track.segments.insert(point_distance.segment_idx + 1, post_segment)

        self._processed_segment_data = {}
        self._processed_track_data = {}


@final
class GPXFileTrack(Track):
    """Track that should be initialized by loading a .gpx file"""

    def __init__(
        self,
        gpx_file: str,
        n_track: int = 0,
        stopped_speed_threshold: float = 1,
        max_speed_percentile: int = 95,
        require_data_extensions: set[str] | None = None,
        heartrate_zones: None | Zones = None,
        power_zones: None | Zones = None,
        cadence_zones: None | Zones = None,
    ) -> None:
        """Initialize a Track object from a gpx file

        :param gpx_file: Path to the gpx file.
        :param n_track: Index of track in the gpx file, defaults to 0
        :param stopped_speed_threshold: Minium speed required for a point to be count
            as moving, defaults to 1
        :param max_speed_percentile: Points with speed outside of the percentile are not
            counted when analyzing the track, defaults to 95
        :param heartrate_zones: Optional heartrate Zones, defaults to None
        :param power_zones: Optional power Zones, defaults to None
        :param cadence_zones: Optional cadence Zones, defaults to None
        """

        super().__init__(
            stopped_speed_threshold=stopped_speed_threshold,
            max_speed_percentile=max_speed_percentile,
            heartrate_zones=heartrate_zones,
            require_data_extensions=require_data_extensions,
            power_zones=power_zones,
            cadence_zones=cadence_zones,
        )

        logger.info("Loading gpx track from file %s", gpx_file)

        gpx = self._get_gpx(gpx_file)

        self._track = gpx.tracks[n_track]
        self._update_extensions()

    @staticmethod
    def _get_gpx(gpx_file: str) -> GPX:
        with open(gpx_file, "r") as f:
            return gpxpy.parse(f)

    @property
    def track(self) -> GPXTrack:
        return self._track


@final
class ByteTrack(Track):
    """Track that should be initialized from a byte stream"""

    def __init__(
        self,
        bytefile: bytes,
        n_track: int = 0,
        stopped_speed_threshold: float = 1,
        max_speed_percentile: int = 95,
        require_data_extensions: set[str] | None = None,
        heartrate_zones: None | Zones = None,
        power_zones: None | Zones = None,
        cadence_zones: None | Zones = None,
    ) -> None:
        """Initialize a Track object from a gpx file

        :param bytefile: Bytestring of a gpx file
        :param n_track: Index of track in the gpx file, defaults to 0
        :param stopped_speed_threshold: Minium speed required for a point to be count
            as moving, defaults to 1
        :param max_speed_percentile: Points with speed outside of the percentile are not
            counted when analyzing the track, defaults to 95
        :param heartrate_zones: Optional heartrate Zones, defaults to None
        :param power_zones: Optional power Zones, defaults to None
        :param cadence_zones: Optional cadence Zones, defaults to None
        """
        super().__init__(
            stopped_speed_threshold=stopped_speed_threshold,
            max_speed_percentile=max_speed_percentile,
            require_data_extensions=require_data_extensions,
            heartrate_zones=heartrate_zones,
            power_zones=power_zones,
            cadence_zones=cadence_zones,
        )

        gpx = gpxpy.parse(bytefile)

        self._track = gpx.tracks[n_track]
        self._update_extensions()

    @property
    def track(self) -> GPXTrack:
        return self._track


@final
class PyTrack(Track):
    """Track that should be initialized from python objects"""

    def __init__(
        self,
        points: list[tuple[float, float]],
        elevations: None | list[float],
        times: None | list[datetime],
        extensions: dict[str, list[N | None] | None] | None = None,
        stopped_speed_threshold: float = 1,
        max_speed_percentile: int = 95,
        require_data_extensions: set[str] | None = None,
        heartrate_zones: None | Zones = None,
        power_zones: None | Zones = None,
        cadence_zones: None | Zones = None,
    ) -> None:
        """A geospacial data track initialized from python objects

        :param points: List of Latitude/Longitude tuples
        :param elevations: Optional list of elevation for each point
        :param times: Optional list of times for each point
        :param heartrate: Optional list of heartrate values for each point
        :param cadence: Optional list of cadence values for each point
        :param power: Optional list of power values for each point
        :param stopped_speed_threshold: Minium speed required for a point to be count
            as moving, defaults to 1
        :param max_speed_percentile: Points with speed outside of the percentile are not
            counted when analyzing the track, defaults to 95
        :param heartrate_zones: Optional heartrate Zones, defaults to None
        :param power_zones: Optional power Zones, defaults to None
        :param cadence_zones: Optional cadence Zones, defaults to None
        :raises TrackInitializationError: Raised if number of elevation, time, heatrate,
            or cadence values do not match passed points
        """
        if extensions is None:
            extensions = dict()

        super().__init__(
            stopped_speed_threshold=stopped_speed_threshold,
            max_speed_percentile=max_speed_percentile,
            require_data_extensions=require_data_extensions,
            heartrate_zones=heartrate_zones,
            power_zones=power_zones,
            cadence_zones=cadence_zones,
            extensions=set(extensions.keys()),
        )

        gpx = GPX()

        gpx_track = GPXTrack()
        gpx.tracks.append(gpx_track)

        gpx_segment = self._create_segmeent(
            points=points,
            elevations=elevations,
            times=times,
            extensions=extensions,
        )

        gpx_track.segments.append(gpx_segment)

        self._track = gpx.tracks[0]

    @property
    def track(self) -> GPXTrack:
        return self._track

    def _create_segmeent(
        self,
        points: list[tuple[float, float]],
        elevations: None | list[float],
        times: None | list[datetime],
        extensions: dict[str, list[N | None] | None],
    ) -> GPXTrackSegment:
        elevations_: list[None] | list[float]
        times_: list[None] | list[datetime]

        if elevations is not None:
            if len(points) != len(elevations):
                raise TrackInitializationError(
                    "Different number of points and elevations was passed"
                )
            elevations_ = elevations
        else:
            elevations_ = len(points) * [None]

        if times is not None:
            if len(points) != len(times):
                raise TrackInitializationError(
                    "Different number of points and times was passed"
                )
            times_ = times
        else:
            times_ = len(points) * [None]

        for key, values in extensions.items():
            if values is None:
                continue
            if len(points) != len(values):
                raise TrackInitializationError(
                    f"Different number of points and {key} was passed"
                )

        gpx_segment = GPXTrackSegment()

        for idx_point in range(len(points)):
            lat, lng = points[idx_point]
            ele = elevations_[idx_point]
            time = times_[idx_point]

            this_extensions = {}
            for key in extensions:
                if extensions[key] is None:
                    this_extensions[key] = None
                else:
                    this_extensions[key] = extensions[key][idx_point]

            this_point = get_extended_track_point(lat, lng, ele, time, this_extensions)

            gpx_segment.points.append(this_point)

        return gpx_segment

    def add_segmeent(  # type: ignore
        self,
        points: list[tuple[float, float]],
        elevations: None | list[float],
        times: None | list[datetime],
        extensions: dict[str, list[N | None] | None] | None = None,
    ) -> None:
        if extensions is None:
            extensions = dict()
        gpx_segment = self._create_segmeent(
            points=points, elevations=elevations, times=times, extensions=extensions
        )
        super().add_segmeent(gpx_segment)


@final
class SegmentTrack(Track):
    """
    Track that should be initialized by loading a PGXTrackSegment object
    """

    def __init__(
        self,
        segment: GPXTrackSegment,
        stopped_speed_threshold: float = 1,
        max_speed_percentile: int = 95,
        require_data_extensions: set[str] | None = None,
        heartrate_zones: None | Zones = None,
        power_zones: None | Zones = None,
        cadence_zones: None | Zones = None,
    ) -> None:
        """Wrap a GPXTrackSegment into a Track object

        :param segment: GPXTrackSegment
        :param stopped_speed_threshold: Minium speed required for a point to be count
            as moving, defaults to 1
        :param max_speed_percentile: Points with speed outside of the percentile are not
            counted when analyzing the track, defaults to 95
        :param heartrate_zones: Optional heartrate Zones, defaults to None
        :param power_zones: Optional power Zones, defaults to None
        :param cadence_zones: Optional cadence Zones, defaults to None
        """
        gpx = GPX()

        gpx_track = GPXTrack()
        gpx.tracks.append(gpx_track)

        gpx_track.segments.append(segment)

        super().__init__(
            stopped_speed_threshold=stopped_speed_threshold,
            max_speed_percentile=max_speed_percentile,
            require_data_extensions=require_data_extensions,
            heartrate_zones=heartrate_zones,
            power_zones=power_zones,
            cadence_zones=cadence_zones,
            extensions=get_extensions_in_points(segment.points),
        )
        self._track = gpx.tracks[0]

    @property
    def track(self) -> GPXTrack:
        return self._track


@final
class FITTrack(Track):
    """Track that should be initialized by loading a .fit file"""

    def __init__(
        self,
        source: str | bytes,
        stopped_speed_threshold: float = 1,
        max_speed_percentile: int = 95,
        strict_elevation_loading: bool = False,
        require_data_extensions: set[str] | None = None,
        heartrate_zones: None | Zones = None,
        power_zones: None | Zones = None,
        cadence_zones: None | Zones = None,
    ) -> None:
        """Load a .fit file and extract the data into a Track object.
        NOTE: Tested with Wahoo devices only

        :param source: Path to fit file or byte representation of fit file
        :param stopped_speed_threshold: Minium speed required for a point to be count
            as moving, defaults to 1
        :param max_speed_percentile: Points with speed outside of the percentile are not
            counted when analyzing the track, defaults to 95
        :param strict_elevation_loading: If set, only points are added to the track that
            have a valid elevation,defaults to False
        :param heartrate_zones: Optional heartrate Zones, defaults to None
        :param power_zones: Optional power Zones, defaults to None
        :param cadence_zones: Optional cadence Zones, defaults to None
        """
        super().__init__(
            stopped_speed_threshold=stopped_speed_threshold,
            max_speed_percentile=max_speed_percentile,
            require_data_extensions=require_data_extensions,
            heartrate_zones=heartrate_zones,
            power_zones=power_zones,
            cadence_zones=cadence_zones,
        )

        if isinstance(source, str):
            logger.info("Loading fit track from file %s", source)
        else:
            logger.info("Using passed bytes data as fit track")

        fit_data = FitFile(
            source,
            data_processor=StandardUnitsDataProcessor(),
        )

        points, elevations, times = [], [], []

        rename_keys = {
            "heart_rate": "heartrate",
            "distance": "raw_distance",
            "speed": "raw_speed",
            "calories": "cum_calories",
        }
        alias_keys = {"enhanced_speed": ["speed"]}
        alias_values = set()
        for value in alias_keys.values():
            alias_values.update(value)

        split_at = set([0])
        extensions = BackFillExtensionDict()
        for record in fit_data.get_messages(("record", "lap")):  # type: ignore
            record: DataMessage  # type: ignore
            if record.mesg_type.name == "lap":
                split_at.add(len(points))
            lat = record.get_value("position_lat")
            long = record.get_value("position_long")
            ele = record.get_value("enhanced_altitude")
            if ele is None and (alt := record.get_value("altitude")) is not None:
                ele = alt
            ts = record.get_value("timestamp")

            check_vals = [lat, long, ts]
            if strict_elevation_loading:
                check_vals.append(ele)

            if any([v is None for v in check_vals]):
                logger.debug(
                    "Found records with None value in lat/long/elevation/timestamp "
                    " - %s/%s/%s/%s",
                    lat,
                    long,
                    ele,
                    ts,
                )
                continue

            record_extensions = {}
            extension_names = []
            for field in record.fields:
                if field.name in [
                    "position_long",
                    "position_lat",
                    "enhanced_altitude",
                    "altitude",
                    "timestamp",
                ]:
                    continue
                extension_names.append(field.name)

            for name in extension_names:
                if name in alias_values:
                    continue
                value = record.get_value(name)
                if name in alias_keys and value is None:
                    for alias in alias_keys[name]:
                        value = record.get_value(alias)
                        if value is not None:
                            break
                record_extensions[rename_keys.get(name, name)] = value

            extensions.fill(record_extensions)

            points.append((lat, long))
            elevations.append(ele)
            times.append(ts)

        if not strict_elevation_loading and set(elevations) != {None}:
            elevations = fill_list(elevations)

        try:
            session_data: DataMessage = list(fit_data.get_messages("session"))[-1]  # type: ignore
        except IndexError:
            logger.debug("Could not load session data from fit file")
        else:
            self.session_data = {  # type: ignore
                "start_time": session_data.get_value("start_time"),
                "ride_time": session_data.get_value("total_timer_time"),
                "total_time": session_data.get_value("total_elapsed_time"),
                "distance": session_data.get_value("total_distance"),
                "ascent": session_data.get_value("total_ascent"),
                "descent": session_data.get_value("total_descent"),
                "avg_velocity": session_data.get_value("avg_speed"),
                "max_velocity": session_data.get_value("max_speed"),
            }

        split_at = sorted(split_at)
        if len(split_at) == 1:
            split_at.append(len(points))

        gpx = GPX()

        gpx_track = GPXTrack()
        gpx.tracks.append(gpx_track)

        for start_idx, end_idx in pairwise(split_at):
            gpx_segment = GPXTrackSegment()

            _points = points[start_idx:end_idx]
            _elevations = elevations[start_idx:end_idx]
            _times = times[start_idx:end_idx]
            _extensions = {
                key: value[start_idx:end_idx] for key, value in extensions.items()
            }

            for i in range(len(_points)):
                this_extensions = {key: _extensions[key][i] for key in _extensions}
                lat, lng = _points[i]
                this_point = get_extended_track_point(
                    lat, lng, _elevations[i], _times[i], this_extensions
                )
                gpx_segment.points.append(this_point)

            gpx_track.segments.append(gpx_segment)

        self._track = gpx.tracks[0]
        self._update_extensions()

    @property
    def track(self) -> GPXTrack:
        return self._track


@final
class GeoJsonTrack(Track):
    """Track that should be initialized by loading a .json file"""

    def __init__(
        self,
        source: str | bytes | dict,
        stopped_speed_threshold: float = 1,
        max_speed_percentile: int = 95,
        allow_empty_spatial: bool = False,
        fallback_coordinates: tuple[float, float] = (0.0, 0.0),
        require_data_extensions: set[str] | None = None,
        heartrate_zones: None | Zones = None,
        power_zones: None | Zones = None,
        cadence_zones: None | Zones = None,
    ) -> None:
        """Load a .json file that conforms to a supported geojson format. Currently
        the GeoJsonTrack supports:
        1. LineString + Arrays
        2. Collection of Points

        :param source: Path or byte representation of the json file
        :param stopped_speed_threshold: Minium speed required for a point to be count
            as moving, defaults to 1
        :param max_speed_percentile: Points with speed outside of the percentile are not
            counted when analyzing the track, defaults to 95
        :param allow_empty_spatial: Allow geojson files without spatial values. If true,
            will generate the track with a fixed location (see. fallback_coordinates)
        :param fallback_coordinates: Coordinates to be used if the geojson file
            does not contain spatial data, defaults to (0.0, 0.0)
        :param heartrate_zones: Optional heartrate Zones, defaults to None
        :param power_zones: Optional power Zones, defaults to None
        :param cadence_zones: Optional cadence Zones, defaults to None
        """
        from geo_track_analyzer.utils.geojson import read_raw_data

        super().__init__(
            stopped_speed_threshold=stopped_speed_threshold,
            max_speed_percentile=max_speed_percentile,
            require_data_extensions=require_data_extensions,
            heartrate_zones=heartrate_zones,
            power_zones=power_zones,
            cadence_zones=cadence_zones,
        )

        if isinstance(source, dict):
            raw_data = source
        elif isinstance(source, str):
            with open(source, "r") as f:
                raw_data = json.load(f)
        else:
            with StringIO(source.decode("utf-8")) as f:
                raw_data = json.load(f)

        _track = read_raw_data(
            data=raw_data,
            allow_empty_spatial=allow_empty_spatial,
            fallback_coordinates=fallback_coordinates,
        )

        self._track = _track.track
        self._update_extensions()

    @property
    def track(self) -> GPXTrack:
        return self._track



import logging
from typing import Callable, Literal

import pandas as pd
import plotly.graph_objects as go
from plotly.graph_objs import Figure
from plotly.subplots import make_subplots

from geo_track_analyzer.exceptions import VisualizationSetupError
from geo_track_analyzer.visualize.metrics import _add_scatter_cont, _add_scatter_disonct
from geo_track_analyzer.visualize.utils import get_slope_colors

logger = logging.getLogger(__name__)


def _check_segment_availability(data: pd.DataFrame) -> bool:
    if "segment" not in data.columns:
        logger.warning(
            "Got show_segments but segments are not part of the data. "
            "Disabling segment borders."
        )
        return False
    if len(data.segment.unique()) == 1:
        logger.warning("Only a single segment in the data. Disabling segment borders.")
        return False
    return True


def _add_segment_borders(data: pd.DataFrame, fig: Figure, color: None | str) -> None:
    for idx, segment_border_idx in enumerate(
        data.index[data["segment"] != data["segment"].shift()].to_list()
    ):
        border_x = data.loc[segment_border_idx].cum_distance_moving
        fig.add_vline(
            x=border_x,
            # Hide line but keep abbitiation for first segment
            line_width=3 if idx > 0 else 0,
            line_dash="dash",
            line_color="darkslategray" if color is None else color,
            annotation_text=f"Segment {data.loc[segment_border_idx].segment}",
            annotation_borderpad=5,
        )


def _add_secondary(
    fig: Figure,
    data: pd.DataFrame,
    secondary: Literal["velocity", "heartrate", "cadence", "power", "speed"],
    split_by_zone: bool,
    min_zone_size: float,
) -> None:
    mode = "lines"
    fill: None | str = "tozeroy"
    y_range_max_factor = 1.2
    y_converter: Callable[[pd.Series], pd.Series] = lambda s: s.fillna(0).astype(int)
    if secondary == "velocity" or secondary == "speed":
        title = "Velocity [km/h]"
        y_range_max_factor = 2.1
        y_converter = lambda s: s * 3.6
        secondary = "speed"
    elif secondary == "heartrate":
        title = "Heart Rate [bpm]"
    elif secondary == "power":
        title = "Power [W]"
    elif secondary == "cadence":
        title = "Cadence [rpm]"
        fill = None
        mode = "markers"
    else:
        raise RuntimeError(
            "%s was passed for secondary. This is not supported" % secondary
        )

    if pd.isna(data[secondary]).all():
        raise VisualizationSetupError(
            f"Requested to plot{secondary} but data is (partially) missing in data"
        )

    if split_by_zone:
        _add_secondary_disonct(
            fig=fig,
            data=data,
            secondary=secondary,
            y_converter=y_converter,
            y_range_max_factor=y_range_max_factor,
            title=title,
            mode=mode,
            fill=fill,
            min_zone_size=min_zone_size,
        )
    else:
        y_data = y_converter(data[secondary])
        y_range = (0, y_data.max() * y_range_max_factor)

        _add_secondary_cont(
            fig=fig,
            x=data.cum_distance_moving,
            y=y_data,
            y_range=y_range,
            title=title,
            mode=mode,
            fill=fill,
        )


def _add_secondary_cont(
    fig: Figure,
    x: pd.Series,
    y: pd.Series,
    y_range: None | tuple[float, float],
    title: str,
    mode: str,
    fill: None | str,
    color: None | str = None,
    legend_group: None | str = None,
    show_legend: bool = False,
    y_title: None | str = None,
) -> None:
    _add_scatter_cont(
        fig=fig,
        x=x,
        y=y,
        y_range=y_range,
        title=title,
        mode=mode,
        fill=fill,
        color=color,
        legend_group=legend_group,
        show_legend=show_legend,
        y_title=y_title,
        is_secondary=True,
    )


def _add_secondary_disonct(
    fig: Figure,
    data: pd.DataFrame,
    secondary: str,
    y_converter: Callable[[pd.Series], pd.Series],
    y_range_max_factor: float,
    title: str,
    mode: str,
    fill: None | str,
    min_zone_size: float,
) -> None:
    _add_scatter_disonct(
        fig=fig,
        data=data,
        metric=secondary,
        y_converter=y_converter,
        y_range_max_factor=y_range_max_factor,
        title=title,
        mode=mode,
        fill=fill,
        min_zone_size=min_zone_size,
        is_secondary=True,
    )


def plot_track_2d(
    data: pd.DataFrame,
    *,
    include_velocity: bool = False,
    include_heartrate: bool = False,
    include_cadence: bool = False,
    include_power: bool = False,
    show_segment_borders: bool = False,
    strict_data_selection: bool = False,
    height: None | int = 600,
    width: None | int = 1800,
    pois: None | list[tuple[float, float]] = None,
    color_elevation: None | str = None,
    color_additional_trace: None | str = None,
    color_poi: None | str = None,
    color_segment_border: None | str = None,
    slider: bool = False,
    split_by_zone: bool = False,
    min_zone_size: float = 0.0025,
    **kwargs,
) -> Figure:
    """Elevation profile of the track. May be enhanced with additional information like
    Velocity, Heartrate, Cadence, and Power.

    :param data: DataFrame containing track data
    :param include_velocity: Plot velocity as second y-axis, defaults to False
    :param include_heartrate: Plot heart rate as second y-axis, defaults to False
    :param include_cadence: Plot cadence as second y-axis, defaults to False
    :param include_power: Plot power as second y-axis, defaults to False
    :param show_segment_borders: If True show vertical lines between segments in track,
        defaults to False. If no segments are present in data, no error is raised.
    :param strict_data_selection: If True only included that passing the minimum speed
        requirements of the Track, defaults to False
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1800
    :param pois: Optional lat/long coordingates to add to the plot as points of
        interest, defaults to None
    :param color_elevation: Color of the elevation as str interpretable by plotly,
        defaults to None
    :param color_additional_trace: Color of velocity/heartrate/cadence/power as str
        interpretable by plotly, defaults to None
    :param color_poi: Color of the pois as str interpretable by plotly, defaults to None
    :param color_segment_border: Color of the segment border lines as str interpretable
        by plotly, defaults to None
    :param slider: Should a slide be included in the plot to zoom into the x-axis,
        defaults to False
    :param split_by_zone: If True, and one for included_* flags is passed and Zones are
        set for the corresponding extension are set, the Zones will be colored
        according to the zone colors.
    :param min_zone_size: Minimum fraction of points required for a distinct zone, if
        split_by_Zone is passed.
    :raises VisualizationSetupError: If more than one of include_velocity,
        include_heartrate, include_cadence, or include_power was set the True
    :raises VisualizationSetupError: If elevation data is missing in the data
    :raises VisualizationSetupError: If the data requried for the additional data is
       missing
    :raises VisualizationSetupError: If split_by_zone is passed but Zones are not set

    :return: Plotly Figure object.
    """
    if split_by_zone:
        color_additional_trace = None
    if (
        sum(
            [
                int(include_velocity),
                int(include_heartrate),
                int(include_cadence),
                int(include_power),
            ]
        )
        > 1
    ):
        raise VisualizationSetupError(
            "Only one of include_velocity, include_heartrate, include_cadence, "
            "and include_power can be set to True"
        )
    mask = data.moving
    if strict_data_selection:
        mask = mask & data.in_speed_percentile

    data_for_plot: pd.DataFrame = data[mask].copy()  # type: ignore

    if show_segment_borders:
        show_segment_borders = _check_segment_availability(data_for_plot)

    if data_for_plot.elevation.isna().all():
        raise VisualizationSetupError("Can not plot profile w/o elevation information")

    fig = make_subplots(specs=[[{"secondary_y": True}]])
    fig.add_trace(
        go.Scatter(
            x=data_for_plot.cum_distance_moving,
            y=data_for_plot.elevation,
            mode="lines",
            name="Elevation [m]",
            fill="tozeroy",
            text=[
                "<b>Lat</b>: {lat:4.6f}°<br><b>Lon</b>: {lon:4.6f}°<br>".format(
                    lon=rcrd["longitude"], lat=rcrd["latitude"]
                )
                for rcrd in data_for_plot.to_dict("records")
            ],
            hovertemplate="<b>Distance</b>: %{x:.1f} km <br><b>Elevation</b>: "
            + "%{y:.1f} m <br>%{text}<extra></extra>",
            showlegend=False,
        ),
        secondary_y=False,
    )
    fig.update_yaxes(
        title_text="Elevation [m]",
        secondary_y=False,
        range=[
            data_for_plot.elevation.min() * 0.97,
            data_for_plot.elevation.max() * 1.05,
        ],
    )
    fig.update_xaxes(title_text="Distance [m]")

    secondary = None
    if include_velocity:
        secondary = "velocity"
    if include_heartrate:
        secondary = "heartrate"
    if include_cadence:
        secondary = "cadence"
    if include_power:
        secondary = "power"

    if secondary is not None:
        _add_secondary(
            fig=fig,
            data=data_for_plot,
            secondary=secondary,
            split_by_zone=split_by_zone,
            min_zone_size=min_zone_size,
        )

    if pois is not None:
        for i_poi, poi in enumerate(pois):
            lat, lng = poi
            poi_data = data_for_plot[
                (data_for_plot.latitude == lat) & (data_for_plot.longitude == lng)
            ]
            if poi_data.empty:
                logger.warning("Could not find POI in data. Skipping")
                continue
            poi_x = poi_data.iloc[0].cum_distance_moving
            poi_y = poi_data.iloc[0].elevation

            fig.add_scatter(
                name=f"PIO {i_poi} @ {lat} / {lng}",
                x=[poi_x],
                y=[poi_y],
                mode="markers",
                marker=dict(
                    size=20,
                    color="MediumPurple" if color_poi is None else color_poi,
                    symbol="triangle-up",
                    standoff=10,
                    angle=180,
                ),
                showlegend=False,
            )

    if show_segment_borders:
        _add_segment_borders(data_for_plot, fig, color_segment_border)

    fig.update_layout(
        showlegend=split_by_zone,
        autosize=False,
        margin={"r": 0, "t": 0, "l": 0, "b": 0},
    )
    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    fig.update_xaxes(
        range=[
            data_for_plot.iloc[0].cum_distance_moving,
            data_for_plot.iloc[-1].cum_distance_moving,
        ]
    )

    if slider:
        fig.update_layout(
            xaxis=dict(
                rangeslider=dict(visible=True),
            )
        )

    if color_elevation is not None:
        fig.data[0].marker.color = color_elevation  # type: ignore
    if color_additional_trace is not None and any(
        [include_velocity, include_heartrate, include_cadence, include_power]
    ):
        fig.data[1].marker.color = color_additional_trace  # type: ignore

    return fig


def plot_track_with_slope(
    data: pd.DataFrame,
    *,
    slope_gradient_color: tuple[str, str, str] = ("#0000FF", "#00FF00", "#FF0000"),
    min_slope: int = -18,
    max_slope: int = 18,
    show_segment_borders: bool = False,
    height: None | int = 600,
    width: None | int = 1800,
    slider: bool = False,
    color_segment_border: None | str = None,
    **kwargs,
) -> Figure:
    """Elevation profile with slopes between points.

    :param data: DataFrame containing track data
    :param slope_gradient_color: Colors for the min, neutral, max slope values,
        defaults to ("#0000FF", "#00FF00", "#FF0000")
    :param min_slope: Minimum slope for the gradient also acts as floor for the
        displayed slope, defaults to -18
    :param max_slope: Maximum  slope for the gradient also acts as ceiling for the
        displayed slope, defaults to 18
    :param show_segment_borders: If True show vertical lines between segments in track,
        defaults to False. If no segments are present in data, no error is raised.
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1800
    :param slider: Should a slide be included in the plot to zoom into the x-axis,
        defaults to False
    :param color_segment_border: Color of the segment border lines as str interpretable
        by plotly, defaults to None
    :raises VisualizationSetupError: If elevation data is missing in the data

    :return: Plotly Figure object
    """
    slope_color_map = get_slope_colors(
        *slope_gradient_color, max_slope=max_slope, min_slope=min_slope
    )

    data = data[data.moving].copy()  # type: ignore

    if data.elevation.isna().all():
        raise VisualizationSetupError("Can not plot profile w/o elevation information")

    if show_segment_borders:
        show_segment_borders = _check_segment_availability(data)

    elevations = data.elevation.to_list()
    diff_elevation = [0]
    for i, elevation in enumerate(elevations[1:]):
        diff_elevation.append(elevation - elevations[i])

    data["elevation_diff"] = diff_elevation

    def calc_slope(row: pd.Series) -> int:
        try:
            slope = round((row.elevation_diff / row.distance) * 100)
        except ZeroDivisionError:
            slope = 0

        if slope > max_slope:
            slope = max_slope
        elif slope < min_slope:
            slope = min_slope

        return slope

    data["slope"] = data.apply(lambda row: calc_slope(row), axis=1).astype(int)

    fig = go.Figure()

    fig.add_trace(
        go.Scatter(
            x=data.cum_distance_moving,
            y=data.elevation,
            mode="lines",
            name="Elevation [m]",
            fill="tozeroy",
        )
    )

    if slider:
        fig.update_layout(
            xaxis=dict(
                rangeslider=dict(visible=True),
            )
        )

    for i in range(len(data)):
        this_data = data.iloc[i : i + 2]
        if len(this_data) == 1:
            continue

        slope_val = this_data.iloc[1].slope

        color = slope_color_map[slope_val]
        max_distance: float = max(this_data.cum_distance_moving)
        fig.add_trace(
            go.Scatter(
                x=this_data.cum_distance_moving,
                y=this_data.elevation,
                mode="lines",
                name=f"Distance {max_distance / 1000:.1f} km",
                fill="tozeroy",
                marker_color=color,
                hovertemplate=f"Slope: {slope_val} %",
            )
        )

    if show_segment_borders:
        _add_segment_borders(data, fig, color_segment_border)

    fig.update_layout(
        showlegend=False, autosize=False, margin={"r": 0, "t": 0, "l": 0, "b": 0}
    )

    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    min_elevation: float = min(data.elevation)
    max_elevation: float = max(data.elevation)
    fig.update_yaxes(
        showspikes=True,
        spikemode="across",
        range=[min_elevation * 0.95, max_elevation * 1.05],
        title_text="Elevation [m]",
    )
    fig.update_xaxes(title_text="Distance [m]")

    return fig



from datetime import datetime
from typing import Annotated, Literal

from pydantic import BaseModel, Field

from geo_track_analyzer import PyTrack, Track
from geo_track_analyzer.exceptions import (
    EmptyGeoJsonError,
    GeoJsonWithoutGeometryError,
    UnsupportedGeoJsonTypeError,
)

Coordinates = tuple[float, float, float]


class LineStringGeometry(BaseModel):
    type: Literal["LineString"] = "LineString"
    coordinates: list[Coordinates]


class PointGeometry(BaseModel):
    type: Literal["Point"] = "Point"
    coordinates: Coordinates


Geometry = Annotated[LineStringGeometry | PointGeometry, Field(discriminator="type")]


class BaseProperties(BaseModel):
    name: str | None = None


class PointProperties(BaseProperties):
    time: datetime
    heartrate: int | None = Field(default=None, alias="heartRate")
    cadence: int | None = None
    power: float | None = None
    temperature: float | None = None


class LineProperties(BaseProperties):
    times: list[datetime] = Field(alias="coordTimes")
    heartrates: list[int | None] | None = Field(default=None, alias="heartRates")
    cadences: list[int | None] | None = None
    powers: list[float | None] | None = None
    temperatures: list[float | None] | None = None


class PointFeature(BaseModel):
    type: Literal["Feature"] = "Feature"
    geometry: PointGeometry | None = None
    properties: PointProperties


class LineFeature(BaseModel):
    type: Literal["Feature"] = "Feature"
    geometry: LineStringGeometry | None = None
    properties: LineProperties


class PointCollection(BaseModel):
    type: Literal["FeatureCollection"] = "FeatureCollection"
    features: list[PointFeature]
    properties: BaseProperties | None = None


class LineCollection(BaseModel):
    type: Literal["FeatureCollection"] = "FeatureCollection"
    features: list[LineFeature]
    properties: BaseProperties | None = None


def _convert_points_collection(
    data: dict, allow_empty_spatial: bool, fallback_coordinates: tuple[float, float]
) -> Track:
    _data = PointCollection.model_validate(data)

    if not allow_empty_spatial and any(g.geometry is None for g in _data.features):
        raise GeoJsonWithoutGeometryError("One or more features have no geometry")

    points = []
    elevations = []
    times = []
    extensions = dict(
        heartrate=[],
        power=[],
        cadence=[],
        temperature=[],
    )

    for feat in _data.features:
        if allow_empty_spatial:
            pass
            lon, lat = fallback_coordinates
            ele = None
        else:
            assert feat.geometry is not None
            lon, lat, ele = feat.geometry.coordinates
        points.append((lat, lon))
        elevations.append(ele)
        times.append(feat.properties.time)
        extensions["heartrate"].append(feat.properties.heartrate)
        extensions["cadence"].append(feat.properties.cadence)
        extensions["power"].append(feat.properties.power)
        extensions["temperature"].append(feat.properties.temperature)

    track = PyTrack(
        points=points,
        elevations=None if allow_empty_spatial else elevations,
        times=times,
        extensions=extensions,  # type: ignore
    )

    return track


def _cast_line_feature_to_data(
    data: LineFeature,
    allow_empty_spatial: bool,
    fallback_coordinates: tuple[float, float],
) -> tuple[
    list[tuple[float, float]], list[float] | None, list[datetime], dict[str, list]
]:
    if not allow_empty_spatial and data.geometry is None:
        raise GeoJsonWithoutGeometryError("Geometry is missing")

    if not allow_empty_spatial and data.geometry is None:
        raise GeoJsonWithoutGeometryError("Geometry is missing")

    _times = data.properties.times
    if allow_empty_spatial:
        _corrdinates = None
    else:
        assert data.geometry is not None
        _corrdinates = data.geometry.coordinates
    _heartrates = data.properties.heartrates
    _cadences = data.properties.cadences
    _powers = data.properties.powers
    _temperatures = data.properties.temperatures

    if _corrdinates is not None and len(_corrdinates) != len(_times):
        raise ValueError("Number of coordinates does not match number of timestamps")
    if _heartrates and len(_times) != len(_heartrates):
        raise ValueError("Number of coordinates does not match number of heartrates")
    if _cadences and len(_times) != len(_cadences):
        raise ValueError("Number of coordinates does not match number of cadences")
    if _powers and len(_times) != len(_powers):
        raise ValueError("Number of coordinates does not match number of powers")
    if _temperatures and len(_times) != len(_temperatures):
        raise ValueError("Number of coordinates does not match number of temperatures")

    points = []
    elevations = []
    if _corrdinates:
        for coord in _corrdinates:
            lon, lat, ele = coord
            points.append((lat, lon))
            elevations.append(ele)
    else:
        elevations = None
        points = [fallback_coordinates] * len(_times)

    return (
        points,
        elevations,
        _times,
        dict(
            heartrate=_heartrates,
            power=_powers,
            cadence=_cadences,
            temperature=_temperatures,
        ),  # type: ignore
    )


def _convert_linestrings_collection(
    data: dict, allow_empty_spatial: bool, fallback_coordinates: tuple[float, float]
) -> Track:
    _data = LineFeature.model_validate(data)

    points, elevations, times, extensions = _cast_line_feature_to_data(
        _data, allow_empty_spatial, fallback_coordinates
    )

    track = PyTrack(
        points=points,
        elevations=elevations,
        times=times,
        extensions=extensions,  # type: ignore
    )

    return track


def _convert_linestrings_multisegment_collection(
    data: dict, allow_empty_spatial: bool, fallback_coordinates: tuple[float, float]
) -> Track:
    _data = LineCollection.model_validate(data)

    if len(_data.features) == 0:
        raise EmptyGeoJsonError("No features in GeoJSON data")

    features_iter = iter(_data.features)

    # Initialize with first segment
    first_segment = next(features_iter)
    points, elevations, times, extensions = _cast_line_feature_to_data(
        first_segment, allow_empty_spatial, fallback_coordinates
    )
    track = PyTrack(
        points=points,
        elevations=elevations,
        times=times,
        extensions=extensions,  # type: ignore
    )

    # Add remaining segments
    for segment in features_iter:
        points, elevations, times, extensions = _cast_line_feature_to_data(
            segment, allow_empty_spatial, fallback_coordinates
        )
        track.add_segmeent(
            points=points,
            elevations=elevations,
            times=times,
            extensions=extensions,  # type: ignore
        )

    return track


def read_raw_data(
    data: dict, allow_empty_spatial: bool, fallback_coordinates: tuple[float, float]
) -> Track:
    format = None
    if data.get("type") == "FeatureCollection":
        features = data.get("features", [])
        if features:
            properties = features[0].get("properties", {})
            if properties.get("coordTimes") is not None:
                format = "linestrings_multisegment_collection"
            else:
                format = "points_collection"
    elif data.get("type") == "Feature":
        format = "linestrings_collection"
    if format is None:
        raise UnsupportedGeoJsonTypeError("Unsupported GeoJSON format")

    if format == "points_collection":
        return _convert_points_collection(
            data=data,
            allow_empty_spatial=allow_empty_spatial,
            fallback_coordinates=fallback_coordinates,
        )
    elif format == "linestrings_collection":
        return _convert_linestrings_collection(
            data=data,
            allow_empty_spatial=allow_empty_spatial,
            fallback_coordinates=fallback_coordinates,
        )
    elif format == "linestrings_multisegment_collection":
        return _convert_linestrings_multisegment_collection(
            data=data,
            allow_empty_spatial=allow_empty_spatial,
            fallback_coordinates=fallback_coordinates,
        )
    else:
        raise ValueError("Unsupported GeoJSON format")
